#ifndef XC_H
#define XC_H

#ifdef WIN32
#define _int16 short
#define _int32 int
#define _int64 __int64
#define _float32 double
#define _float64 long double

#define xc_get_errno() GetLastError()
#define xc_set_errno(X) SetLastError((X))

#else
#define _int16 short
#define _int32 int
#define _int64 long long
#define _float32 float
#define _float64 double

#define xc_get_errno() (errno)
#define xc_set_errno(X) (errno = (X))

#endif

#define _uint16 unsigned _int16
#define _uint32 unsigned _int32
#define _uint64 unsigned _int64

#ifdef WIN32
	#include <windows.h>
	#include <winsock.h>
	#include <io.h>
#else
	#include <sys/types.h>
	#include <sys/socket.h>
	#include <sys/time.h>
	#include <unistd.h>
	#include <netdb.h>

	struct in_addr
	{
        unsigned long s_addr;
	};

	typedef struct sockaddr SOCKADDR;

	typedef struct sockaddr_in
	{
        short   sin_family;
        u_short sin_port;
        struct  in_addr sin_addr;
        char    sin_zero[8];
	} SOCKADDR_IN;
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <fcntl.h>
#include <time.h>

#define SIZE_8 0
#define SIZE_16 1
#define SIZE_32 2
#define SIZE_64 3

#define xc_func(X) _int64 X(xinst *_xinst, int argc, _int64 argv[])

#define XC_ERR_NONE 0
#define XC_ERR_FILE_OPEN 1
#define XC_ERR_MALLOC 2
#define XC_ERR_INST_LIMIT 3
#define XC_ERR_NO_INST 4
#define XC_ERR_SEGFAULT 5
#define XC_ERR_DIV_BY_ZERO 6
#define XC_ERR_STACK_OVERFLOW 7
#define XC_ERR_STACK_UNDERFLOW 8
#define XC_ERR_CSTACK_OVERFLOW 9
#define XC_ERR_HOSTSYM_OVERFLOW 10
#define XC_ERR_HOSTSYM_SEGFAULT 11
#define XC_ERR_HOSTSYM_TYPE 12
#define XC_ERR_NULL_PTR 13
#define XC_ERR_MODE_OVERFLOW 14
#define XC_ERR_MODE_UNDERFLOW 15
#define XC_ERR_MODULE_MEMORY 16
#define XC_ERR_NOT_EXECUTING 17
#define XC_ERR_INVALID_OP 18

#define XC_TYPE_VOID 0
#define XC_TYPE_CHAR 1
#define	XC_TYPE_INT 2
#define XC_TYPE_SHORT 4
#define XC_TYPE_LONG 5

#define XN_TCP 1
#define XN_UDP 2
#define XN_TCP_BLOCKING 3
#define XN_TYPEMAX 3

#define XN_ERR_BAD_TYPE 1
#define XN_ERR_CREATE 2
#define XN_ERR_LIBRARIES 3
#define XN_ERR_BIND 4
#define XN_ERR_LISTEN 5
#define XN_ERR_UNSUPPORTED 6
#define XN_ERR_TYPE_COMPAT 7
#define XN_ERR_ACCEPT 8
#define XN_ERR_SEND 9
#define XN_ERR_RECV 10
#define XN_ERR_CONNECT 11
#define XN_ERR_CLOSED 12
#define XN_ERR_BLOCKING 16
#define XN_ERR_RESOLVE 17

#define XC_NUM_IVARS 32

#define XN_DISCON 0
#define XN_LISTEN 1
#define XN_CON 2
#define XN_WAIT 3
#define XN_TRANS 4
#define XN_ALTER 5

typedef _int64 (*xcfunc)();
typedef struct xc xc;
typedef struct xinst xinst;

typedef struct tcp_info
{
	SOCKADDR_IN local, remote;
	int sock;
	short lport, rport;
} tcp_info;

typedef union xnet_info
{
	tcp_info tcp;
} xnet_info;

typedef struct xnet
{
	int type, mode, lmode;
	xnet_info info;
} xnet;

#ifdef __cplusplus
extern "C" {
#endif

int xc_export_func(xcfunc);
#ifdef HOST_TABLE
	extern n_hostfunc;
	#define xc_export_func(X) (printf("[%d] " #X "\n", n_hostfunc), xc_export_func((X)))
#endif

xc *load_xc(char*);
xc *load_xc_string(char*, int);
xc *new_xc_stream(int);
int unload_xc(xc*);
int abort_xc(int);
int abort_all_xc();
void xc_callback(void (*f)(void *));
void xc_callback_arg(void *a);
int run_xc(xc*, unsigned int, void*, xnet*, xnet*, xnet*, int, ...);
int unload_xinst(xinst*);

int xnet_init();
int xnet_create(xnet*, int);
int xnet_connect(xnet*, ...);
int xnet_listen(xnet*, ...);
int xnet_accept(xnet*, xnet*);
int xnet_close(xnet*);
int xnet_send(xnet*, char*, int);
int xnet_sends(xnet*, char*);
int xnet_sendc(xnet*, char);
int xnet_recv(xnet*, char*, int);
int xnet_recvs(xnet*, char*, int);
int xnet_newdata(xnet*);
int xnet_cleanup();

#define xc_malloc(X) _xc_malloc(_xinst, (X))
void *_xc_malloc(xinst *, int);
#define xc_free(X) _xc_free(_xinst, (X))
int _xc_free(xinst *, void *);
#define xc_block_size(X) _xc_block_size(_xinst, (X))
int _xc_block_size(xinst *, void *);
#define xc_ptr_ok(X) _xc_ptr_ok(_xinst, (X))
int _xc_ptr_ok(xinst *, void *);

char xc_char_in(_int64);
_int16 xc_int16_in(_int64);
_int32 xc_int32_in(_int64);
_int64 xc_int64_in(_int64);
_float32 xc_float32_in(_int64);
_float64 xc_float64_in(_int64);

#define xc_ptr_in(X, Y) _xc_ptr_in(_xinst, (X), (Y))
void *_xc_ptr_in(xinst *, _int64, int);
#define xc_string_in(X) _xc_string_in(_xinst, (X))
char *_xc_string_in(xinst *, _int64);
#define xc_fptr_in(X) _xc_fptr_in(_xinst, (X))
_int32 _xc_fptr_in(xinst *, _int64 val);

_int64 xc_char_out(char);
_int64 xc_int16_out(_int16);
_int64 xc_int32_out(_int32);
_int64 xc_int64_out(_int64);
_int64 xc_float32_out(_float32);
_int64 xc_float64_out(_float64);

#define xc_ptr_out(X) _xc_ptr_out(_xinst, (X))
_int64 _xc_ptr_out(xinst *, void *);
int xc_ping(xnet *x, time_t to);

#define xc_mem_write(X, Y, Z) _xc_mem_write(_xinst, (X), (Y), (Z))
void _xc_mem_write(xinst *, unsigned char *ptr, _int64 val, int sz);
#define xc_mem_read(X, Y) _xc_mem_read(_xinst, (X), (Y))
_int64 _xc_mem_read(xinst *, unsigned char *ptr, int sz);
_int64 xc_func_call(xinst *, _int32 loc, int argc, ...);
#define xc_atexit(X) _xc_atexit(_xinst, (X))
int _xc_atexit(xinst *, _int32 loc);
#define xc_set_ivar(X, Y) _xc_set_ivar(_xinst, (X), (Y))
int _xc_set_ivar(xinst *, int n, _int32 val);
#define xc_get_ivar(X) _xc_get_ivar(_xinst, (X))
_int32 _xc_get_ivar(xinst *, int n);
void *xc_get_tag(xinst *x);
void xc_set_tag(xinst *x, void *t);

#ifdef __cplusplus
}
#endif

#endif
