#include "xc.h"
#include "xctypes.h"

int mantissa[4] = {0, 0, 23, 52};
#ifdef WIN32
_int64 mmask[4] = {0, 0, 0x7FFFFF, 0x0FFFFFFFFFFFFF};
#else
_int64 mmask[4] = {0, 0, 0x7FFFFF, 0x0FFFFFFFFFFFFFULL};
#endif
int exponent[4] = {0, 0, 8, 11};
int bias[4] = {0, 0, 127, 1023};
char tsz[4] = {1, 2, 4, 8};
char cull[4] = {0, 0, 12, 26};
int divbump1[4] = {0, 0, 1024, 4194304};
int divbump2[4] = {0, 0, 10, 22};
int fullexp[4] = {0, 0, 0xFF, 0x7FF};
#ifdef WIN32
unsigned _int64 msb[4] = {0x80, 0x8000, 0x80000000, 0x8000000000000000};
unsigned _int64 imax[4] = {0xFF, 0xFFFF, 0xFFFFFFFF, 0xFFFFFFFFFFFFFFFF};
unsigned _int64 makeneg[4] = {0xFFFFFFFFFFFFFF00, 0xFFFFFFFFFFFF0000, 0xFFFFFFFF00000000, 0x0};
#else
unsigned _int64 msb[4] = {0x80ULL, 0x8000ULL, 0x80000000ULL, 0x8000000000000000ULL};
unsigned _int64 imax[4] = {0xFFULL, 0xFFFFULL, 0xFFFFFFFFULL, 0xFFFFFFFFFFFFFFFFULL};
unsigned _int64 makeneg[4] = {0xFFFFFFFFFFFFFF00ULL, 0xFFFFFFFFFFFF0000ULL, 0xFFFFFFFF00000000ULL, 0x0ULL};
#endif

int opsz[N_OPS] = {

1, // OP_NOP (0)
2, // OP_LD_RR (1)
2, // OP_LD_RR #0 (2)
2, // OP_LD_RR #1 (3)
2, // OP_LD_RR #2 (4)
3, // OP_LD_RC (5)
4, // OP_LD_RC #0 (6)
6, // OP_LD_RC #1 (7)
10, // OP_LD_RC #2 (8)
6, // OP_LD_RMC (9)
6, // OP_LD_RMC #0 (10)
6, // OP_LD_RMC #1 (11)
6, // OP_LD_RMC #2 (12)
2, // OP_LD_RMR (13)
2, // OP_LD_RMR #0 (14)
2, // OP_LD_RMR #1 (15)
2, // OP_LD_RMR #2 (16)
4, // OP_LD_RHC (17)
4, // OP_LD_RHC #0 (18)
4, // OP_LD_RHC #1 (19)
4, // OP_LD_RHC #2 (20)
2, // OP_LD_RHR (21)
2, // OP_LD_RHR #0 (22)
2, // OP_LD_RHR #1 (23)
2, // OP_LD_RHR #2 (24)
6, // OP_LD_MCR (25)
6, // OP_LD_MCR #0 (26)
6, // OP_LD_MCR #1 (27)
6, // OP_LD_MCR #2 (28)
2, // OP_LD_MRR (29)
2, // OP_LD_MRR #0 (30)
2, // OP_LD_MRR #1 (31)
2, // OP_LD_MRR #2 (32)
4, // OP_LD_HCR (33)
4, // OP_LD_HCR #0 (34)
4, // OP_LD_HCR #1 (35)
4, // OP_LD_HCR #2 (36)
2, // OP_LD_HRR (37)
2, // OP_LD_HRR #0 (38)
2, // OP_LD_HRR #1 (39)
2, // OP_LD_HRR #2 (40)
2, // OP_ADD (41)
2, // OP_ADD #0 (42)
2, // OP_ADD #1 (43)
2, // OP_ADD #2 (44)
2, // OP_FADD (45)
2, // OP_FADD #0 (46)
2, // OP_FADD #1 (47)
2, // OP_FADD #2 (48)
2, // OP_SUB (49)
2, // OP_SUB #0 (50)
2, // OP_SUB #1 (51)
2, // OP_SUB #2 (52)
2, // OP_FSUB (53)
2, // OP_FSUB #0 (54)
2, // OP_FSUB #1 (55)
2, // OP_FSUB #2 (56)
2, // OP_MULT (57)
2, // OP_MULT #0 (58)
2, // OP_MULT #1 (59)
2, // OP_MULT #2 (60)
2, // OP_MULTS (61)
2, // OP_MULTS #0 (62)
2, // OP_MULTS #1 (63)
2, // OP_MULTS #2 (64)
2, // OP_FMULT (65)
2, // OP_FMULT #0 (66)
2, // OP_FMULT #1 (67)
2, // OP_FMULT #2 (68)
2, // OP_TEST (69)
2, // OP_TEST #0 (70)
2, // OP_TEST #1 (71)
2, // OP_TEST #2 (72)
2, // OP_DIV (73)
2, // OP_DIV #0 (74)
2, // OP_DIV #1 (75)
2, // OP_DIV #2 (76)
2, // OP_DIVS (77)
2, // OP_DIVS #0 (78)
2, // OP_DIVS #1 (79)
2, // OP_DIVS #2 (80)
2, // OP_FDIV (81)
2, // OP_FDIV #0 (82)
2, // OP_FDIV #1 (83)
2, // OP_FDIV #2 (84)
2, // OP_FCP (85)
2, // OP_FCP #0 (86)
2, // OP_FCP #1 (87)
2, // OP_FCP #2 (88)
2, // OP_MOD (89)
2, // OP_MOD #0 (90)
2, // OP_MOD #1 (91)
2, // OP_MOD #2 (92)
2, // OP_MODS (93)
2, // OP_MODS #0 (94)
2, // OP_MODS #1 (95)
2, // OP_MODS #2 (96)
2, // OP_SL (97)
2, // OP_SL #0 (98)
2, // OP_SL #1 (99)
2, // OP_SL #2 (100)
2, // OP_SLN (101)
2, // OP_SLN #0 (102)
2, // OP_SLN #1 (103)
2, // OP_SLN #2 (104)
2, // OP_SR (105)
2, // OP_SR #0 (106)
2, // OP_SR #1 (107)
2, // OP_SR #2 (108)
2, // OP_SRN (109)
2, // OP_SRN #0 (110)
2, // OP_SRN #1 (111)
2, // OP_SRN #2 (112)
2, // OP_NEG (113)
2, // OP_NEG #0 (114)
2, // OP_NEG #1 (115)
2, // OP_NEG #2 (116)
2, // OP_FNEG (117)
2, // OP_FNEG #0 (118)
2, // OP_FNEG #1 (119)
2, // OP_FNEG #2 (120)
2, // OP_OR (121)
2, // OP_OR #0 (122)
2, // OP_OR #1 (123)
2, // OP_OR #2 (124)
2, // OP_AND (125)
2, // OP_AND #0 (126)
2, // OP_AND #1 (127)
2, // OP_AND #2 (128)
2, // OP_NOT (129)
2, // OP_NOT #0 (130)
2, // OP_NOT #1 (131)
2, // OP_NOT #2 (132)
2, // OP_XOR (133)
2, // OP_XOR #0 (134)
2, // OP_XOR #1 (135)
2, // OP_XOR #2 (136)
2, // OP_BOR (137)
2, // OP_BOR #0 (138)
2, // OP_BOR #1 (139)
2, // OP_BOR #2 (140)
2, // OP_BAND (141)
2, // OP_BAND #0 (142)
2, // OP_BAND #1 (143)
2, // OP_BAND #2 (144)
2, // OP_BNOT (145)
2, // OP_BNOT #0 (146)
2, // OP_BNOT #1 (147)
2, // OP_BNOT #2 (148)
2, // OP_CP (149)
2, // OP_CP #0 (150)
2, // OP_CP #1 (151)
2, // OP_CP #2 (152)
5, // OP_JP (153)
5, // OP_JP_Z (154)
5, // OP_JP_NZ (155)
5, // OP_JP_C (156)
5, // OP_JP_NC (157)
5, // OP_CALL_C (158)
2, // OP_CALL_R (159)
1, // OP_RET (160)
2, // OP_RETV (161)
2, // OP_RETV #0 (162)
2, // OP_RETV #1 (163)
2, // OP_RETV #2 (164)
3, // OP_HCALL_C (165)
2, // OP_HCALL_R (166)
2, // OP_INC (167)
2, // OP_INC #0 (168)
2, // OP_INC #1 (169)
2, // OP_INC #2 (170)
2, // OP_DEC (171)
2, // OP_DEC #0 (172)
2, // OP_DEC #1 (173)
2, // OP_DEC #2 (174)
2, // OP_PUSH (175)
2, // OP_PUSH #0 (176)
2, // OP_PUSH #1 (177)
2, // OP_PUSH #2 (178)
2, // OP_POP (179)
2, // OP_POP #0 (180)
2, // OP_POP #1 (181)
2, // OP_POP #2 (182)
2, // OP_JP_R (183)
2, // OP_JP_R_Z (184)
2, // OP_JP_R_NZ (185)
2, // OP_JP_R_C (186)
2, // OP_JP_R_NC (187)
2, // OP_ARG_C (188)
3, // OP_ARG_C #0 (189)
5, // OP_ARG_C #1 (190)
9, // OP_ARG_C #2 (191)
2, // OP_ARG_R (192)
2, // OP_ARG_R #0 (193)
2, // OP_ARG_R #1 (194)
2, // OP_ARG_R #2 (195)
2, // OP_HCPY (196)
2, // OP_HSCPY (197)
2, // OP_MCPY (198)
2, // OP_CPY (199)
2, // OP_SCPY (200)
2, // OP_DCPY (201)
1, // OP_PMODE (202)
1, // OP_IMODE (203)
1, // OP_RMODE (204)
1, // OP_OMODE (205)
1, // OP_EMODE (206)
1, // OP_SYNC (207)
2, // OP_RSEND (208)
2, // OP_MSEND (209)
5, // OP_MBASE (210)
2, // OP_MALLOC (211)
2, // OP_FREE (212)
5, // OP_LALLOC (213)
1, // OP_LFREE (214)
5, // OP_LADDR (215)
2, // OP_SCP (216)
2, // OP_SCP (217)
2, // OP_SCP (218)
2, // OP_SCP (219)
2, // OP_STEST (220)
2, // OP_STEST (221)
2, // OP_STEST (222)
2, // OP_STEST (223)
2, // OP_SLEN (224)
2, // OP_FNEG (225)
2, // OP_FNEG (226)
2, // OP_FNEG (227)
5, // OP_CAST (228)
2, // OP_FINC (229)
2, // OP_FINC (230)
2, // OP_FINC (231)
2, // OP_FINC (232)
2, // OP_FDEC (233)
2, // OP_FDEC (234)
2, // OP_FDEC (235)
2, // OP_FDEC (236)
4, // OP_GBF (237)
4, // OP_SBF (238)
4, // OP_SGBF (239)
4, // OP_SSBF (240)
2, // OP_ZERO (241)
1, // OP_PING (242)
1, // OP_PONG (243)
};

//hostsymt hostsym[N_HOSTSYM];
//int n_hostsym;
xcfunc hostfunc[N_HOSTSYM];
int n_hostfunc, _abort_flag = 0, _abnormal_flag = 0;

xinst *insts = NULL, *c_inst_p = NULL, *mono_p = NULL;