// xCode interpreter

#ifdef WIN32
	#include <windows.h>
	#include <winsock.h>
	#define sleep(X) Sleep(X)
	#define YIELD Sleep(1)
#endif

#define YCOUNT 1000

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "xc.h"
#include "xctypes.h"
#include "xcvars.h"

#ifndef WIN32
	#include <pthread.h>
	#define YIELD sched_yield()
#endif

/*void write_msg(char*);
void msg_printf(char*, ...);
void log_printf(char*, ...);
#define SPREAD log_printf("Spread: %u %u %u %u [%u] %u %u %u %u\n",	(*cp)[-4], (*cp)[-3], (*cp)[-2], (*cp)[-1], (*cp)[0], (*cp)[1], (*cp)[2], (*cp)[3], (*cp)[4])*/

static do_xinst(xinst *x);
static do_xinst_net(xinst *x);
static do_xinst_streamed(xinst *x);

#define BS 51200 // That's a 50k limit

#define _C_SIGNED 1
#define _C_UNSIGNED 2
#define _C_FLOAT 3

// Begin xCode section

void (*cur_callback)(void *) = NULL, *cur_arg = NULL;
int n_insts = 0;
char active_inst;

int xc_export_func(xcfunc ptr)
{
	if (n_hostfunc >= N_HOSTSYM)
	{
		xc_set_errno(XC_ERR_HOSTSYM_OVERFLOW);
		return 0;
	}

	hostfunc[n_hostfunc] = ptr;

	return n_hostfunc++;
}

static _int64 call_hostsym(int num, xinst *xi)
{
	_int64 ret;

	xc_set_errno(XC_ERR_NONE);

	ret = hostfunc[num](xi, xi->narg, xi->arg);
	return ret;
}

xc *load_xc(char *path)
{
	xc *x;
	int ch1;
	unsigned int j, sz;
	char *cp;
	unsigned char *sp;
	FILE *fl;
	struct stat s;

	if (stat(path, &s) == -1)
	{
		xc_set_errno(XC_ERR_FILE_OPEN);
		return NULL;
	}
	if (!(fl = fopen(path, "rb")))
	{
		xc_set_errno(XC_ERR_FILE_OPEN);
		return NULL;
	}

	if (!(x = (xc *)calloc(1, sizeof(xc))))
	{
		fclose(fl);
		xc_set_errno(XC_ERR_MALLOC);
		return NULL;
	}
	if (!(cp = x->code = (char *)malloc(s.st_size)))
	{
		fclose(fl);
		free(x);
		xc_set_errno(XC_ERR_MALLOC);
		return NULL;
	}

	x->n_data = 0;
	while (1)
	{
		sz = fgetc(fl);
		sz += fgetc(fl) << 8;
		sz += fgetc(fl) << 16;
		sz += fgetc(fl) << 24;
		if (!sz) break;

		x->data = (datat *)realloc(x->data, sizeof(datat) * (x->n_data+1));
		x->data[x->n_data].sz = sz;
		x->data[x->n_data].dat = (char *)calloc(1, sz);
		sp = x->data[x->n_data++].dat;
		for (j = 0; j < sz; ++j)
			*sp++ = fgetc(fl);
	}

	while ((ch1 = fgetc(fl)) != EOF)
		*cp++ = (char)ch1;
	*cp = '\0';
	fclose(fl);

	x->code_end = x->code + s.st_size + 1;

	x->sz = s.st_size;
	x->mode = XC_LOCAL;

	return x;
}

xc *load_xc_string(char *str, int csz)
{
	xc *x;
	char code[BS], *cp = code;

	if (!(x = (xc *)calloc(1, sizeof(xc))))
	{
		xc_set_errno(XC_ERR_MALLOC);
		return NULL;
	}

	x->n_data = 0;

	if (!(x->code = (char *)malloc(csz)))
	{
		free(x);
		xc_set_errno(XC_ERR_MALLOC);
		return NULL;
	}
	x->code_end = x->code + csz + 1;

	memcpy(x->code, str, csz);
	x->sz = csz;
	x->mode = XC_LOCAL;

	return x;
}

int unload_xc(xc *x)
{
	int j;
	xinst *xi;

	for (xi = insts; xi; xi = xi->next)
		if (xi->parent == x)
			unload_xinst(xi);

	if (x->data)
	{
		for (j = 0; j < x->n_data; ++j)
			free(x->data[j].dat);
		free(x->data);
	}

	if (x->code)
		free(x->code);
	free(x);

	return 1;
}

int abort_xc(int abnorm)
{
	_abort_flag = 1;
	_abnormal_flag = abnorm;
	return 1;
}

int abort_all_xc()
{
	xinst *xi, *xi2;

	for (xi = insts; xi; xi = xi2)
	{
		xi2 = xi->next;
		unload_xinst(xi);
	}

	return 1;
}

xc *new_xc_stream(int sz)
{
	xc *x;

	if (!(x = (xc *)calloc(1, sizeof(xc))))
	{
		xc_set_errno(XC_ERR_MALLOC);
		return NULL;
	}

	x->mode = XC_STREAMED;
	x->sz = sz;

	return x;
}

void xc_callback(void (*f)(void *))
{
	cur_callback = f;
}

void xc_callback_arg(void *a)
{
	cur_arg = a;
}

int run_xc(xc *x, unsigned int memsz, void *tag, xnet *in, xnet *out, xnet *serv, int narg, ...)
{
	int j, k;
	xinst *xi;
	char *mem;
	va_list ap;
	_int64 i;
	unsigned char *s_code;
	stackt *s;
	time_t t;

	if (!x)
	{
		xc_set_errno(XC_ERR_NULL_PTR);
		return 0;
	}

	xc_set_errno(0);
	if (!(mem = (unsigned char *)malloc(memsz)))
	{
		xc_set_errno(XC_ERR_MALLOC);
		return 0;
	}

	if (!(xi = (xinst *)calloc(1, sizeof(xinst))))
	{
		free(mem);
		xc_set_errno(XC_ERR_MALLOC);
		return 0;
	}

	memset(xi, 0, sizeof(xinst));
	xi->parent = x;
	xi->code = x->code;
	xi->callback = cur_callback;
	xi->callback_arg = cur_arg;
	xi->tag = tag;

	if (serv)
	{
		if (x->sz)
			xi->sz = x->sz;
		else
			xi->sz = S_NBUF;
		if (!(s_code = (char *)malloc(xi->sz)))
		{
			free(mem);
			free(xi);
			xc_set_errno(XC_ERR_MALLOC);
			return 0;
		}
		xi->nstat = XN_CON;
		xi->s_code_end = (xi->s_code = s_code) - 1;
	}
	if (x->mode != XC_STREAMED)
	{
		xi->cp = x->code;
		xi->nstat = XN_DISCON;
	}

	xi->mem = mem;
	xi->memsz = memsz;
	xi->code_end = x->code_end;
	xi->arg = malloc(0);
	xi->priority = 255;
	memset(xi->reg, 0, N_REG * S_REG);
	xi->in = in;
	xi->out = out;
	if (xi->serv = serv)
		serv->mode = XN_CON;
	xi->mode = OP_PMODE;

	if (narg)
	{
		_int64 *arr = calloc(narg, sizeof(_int64));
		va_start(ap, narg);
		for (j = 0; j < narg; ++j)
			arr[narg - j - 1] = va_arg(ap, _int64);
		for (j = 0; j < narg; ++j)
		{
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = xi->stack;
			mem = s->n;
			i = arr[j];
			for (k = 0; k < S_REG; ++k)
			{
				*mem++ = (char)(i & 255);
				i >>= 8;
			}
			xi->stack = s;
		}
		va_end(ap);
		free(arr);
	}

	t = time(0);
	if (!xi->code)
		do_xinst_streamed(xi);
	else if (in || out)
		do_xinst_net(xi);
	else
		do_xinst(xi);

	j = xc_get_errno();
	unload_xinst(xi);
	xc_set_errno(j);

//	msg_printf("Time to run: %d", (int)(time(0) - t));

	return 1;
}

int unload_xinst(xinst *x)
{
	char f = 0;
	memblock *m, *m2;
	lint *n, *n2;
	stackt *s, *s2;
	pstack *p, *p2;
	stack8 *s8, *s82;

	for (; x->atex; x->atex = n)
	{
		n = x->atex->next;
		if (!_abnormal_flag)
			xc_func_call(x, x->atex->n, 0);
		free(x->atex);
	}
	_abnormal_flag = 0;

	free(x->mem);
	if (x->s_code) free(x->s_code);

	for (m = x->mb_base; m; m = m2)
	{
		m2 = m->next;
		free(m);
	}
	for (n = x->local; n; n = n2)
	{
		n2 = n->next;
		free(n);
	}
	for (s = x->stack; s; s = s2)
	{
		s2 = s->next;
		free(s);
	}
	for (p = x->cstack; p; p = p2)
	{
		p2 = p->next;
		free(p);
	}
	for (s8 = x->cmode; s8; s8 = s82)
	{
		s82 = s8->next;
		free(s8);
	}
	free(x->arg);

	free(x);

	--n_insts;
	return 1;
}

int __xc_malloc(xinst *x, int i1)
{
	memblock *mb, *mb2;

	if (i1 < 0 || i1 >= (int)x->memsz || !x->mb_base)
	{
		xc_set_errno(XC_ERR_MODULE_MEMORY);
		return -1;
	}

	for (mb = x->mb_base; mb; mb = mb->next)
	{
		if (mb->next)
		{
			if (mb->next->min - mb->max - 1 >= i1)
			{
				if (!(mb2 = (memblock *)calloc(1, sizeof(memblock))))
				{
					xc_set_errno(XC_ERR_MODULE_MEMORY);
					return -1;
				}

				mb2->prev = mb;
				mb2->next = mb->next;
				mb->next->prev = mb2;
				mb->next = mb2;

				mb2->min = mb->max + 1;
				mb2->max = mb->max + i1;

				xc_set_errno(XC_ERR_NONE);
				return mb2->min;
			}
		}
		else
		{
			if ((int)x->memsz - mb->max <= i1)
			{
				xc_set_errno(XC_ERR_MODULE_MEMORY);
				return -1;
			}

			if (!(mb2 = (memblock *)calloc(1, sizeof(memblock))))
			{
				xc_set_errno(XC_ERR_MODULE_MEMORY);
				return -1;
			}
			mb2->prev = mb;
			mb->next = mb2;

			mb2->min = mb->max + 1;
			mb2->max = mb->max + i1;

			xc_set_errno(XC_ERR_NONE);
			return mb2->min;
		}
	}

	xc_set_errno(XC_ERR_MODULE_MEMORY);
	return -1;
}

void __xc_free(xinst *x, int i1)
{
	memblock *mb;

	if (i1 < 0 || i1 >= (int)x->memsz || !x->mb_base)
	{
		xc_set_errno(XC_ERR_MODULE_MEMORY);
		return;
	}

	for (mb = x->mb_base->next; mb; mb = mb->next)
	{
		if (i1 >= mb->min && i1 <= mb->max)
		{
			mb->prev->next = mb->next;
			if (mb->next)
				mb->next->prev = mb->prev;
			free(mb);
			return;
		}
	}

	xc_set_errno(XC_ERR_MODULE_MEMORY);
	return;
}

static do_recv(xinst *x)
{
	int n, e;

	if (x->nbufp >= x->transz)
	{
		memcpy(x->transp, x->nbuf, x->transz);
		memmove(x->nbuf, &x->nbuf[x->transz], x->nbufp - x->transz);
		x->nbufp -= x->transz;
		return 1;
	}
	else if (x->nbufp)
	{
		memcpy(x->transp, x->nbuf, x->nbufp);
		x->transz -= x->nbufp;
		x->transp += x->nbufp;
		x->nbufp = 0;
	}

	while (x->transz)
	{
		n = xnet_recv(x->serv, x->transp, x->transz);
		if (e = xc_get_errno())
			if (e != XN_ERR_BLOCKING)
				return 0;
		if (n > 0)
		{
			x->transz -= n;
			x->transp += n;
		}
	}

	return 1;
}

static do_op(xinst *x, unsigned char **cp, unsigned char *code)
{
	unsigned ku;
	unsigned char r1, r2;
	unsigned _int64 i1;
	memblock *mb;
	lint *ln;
	stackt *s;
	stack8 *s8;
	pstack *p;
	int k;

#ifdef WEIRD
	double db1, db2;
	int osz;
	unsigned char *sp1, *sp2, ch, r3;
	signed _int64 s1, s2, mn1, mn2;
	unsigned _int32 ui1, ui2;
	unsigned _int64 i2, i3, im;
	signed _int32 ep1, ep2;
	int narg;
	_int64 *arg;
#endif

#ifndef WEIRD
	unsigned char i8_1, i8_2;
	_uint16 i16_1, i16_2;
	_uint32 i32_1, i32_2, i32_3;
	_uint64 i64_1, i64_2;
	char s8_1, s8_2;
	_int16 s16_1, s16_2;
	_int32 s32_1, s32_2;
	_int64 s64_1, s64_2;
	_float32 f32_1, f32_2;
	_float64 f64_1, f64_2;
#endif

//	log_printf("OP: %d (%c)\n", **cp, x->mem[1055]);

	switch (**cp)
	{
#ifdef WEIRD
		case OP_LD_RR:
		case OP_LD_RR+1:
		case OP_LD_RR+2:
		case OP_LD_RR+3:
			osz = tsz[**cp - OP_LD_RR];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			memcpy(x->reg[r1], x->reg[r2], osz);
			break;
		case OP_LD_RC:
		case OP_LD_RC+1:
		case OP_LD_RC+2:
		case OP_LD_RC+3:
			osz = tsz[**cp - OP_LD_RC];
			(*cp)++;
			r1 = **cp & 15;

			for (k = 0; k < osz; ++k)
				x->reg[r1][k] = *++*cp;

			break;
		case OP_LD_RMC:
		case OP_LD_RMC+1:
		case OP_LD_RMC+2:
		case OP_LD_RMC+3:
			osz = tsz[**cp - OP_LD_RMC];
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*++*cp << (8 * k);
			if (i1+ osz > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			memcpy(x->reg[r1], &(x->mem[i1]), osz);
			break;
		case OP_LD_RMR:
		case OP_LD_RMR+1:
		case OP_LD_RMR+2:
		case OP_LD_RMR+3:
			osz = tsz[**cp - OP_LD_RMR];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r2];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			if (i1 + osz > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			memcpy(x->reg[r1], &(x->mem[i1]), osz);
			break;
		case OP_LD_MCR:
		case OP_LD_MCR+1:
		case OP_LD_MCR+2:
		case OP_LD_MCR+3:
			osz = tsz[**cp - OP_LD_MCR];
			i1 = 0;
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*++*cp << (8 * k);
			r2 = *++*cp & 15;
			if (i1+ osz > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			memcpy(&(x->mem[i1]), x->reg[r2], osz);
			break;
		case OP_LD_MRR:
		case OP_LD_MRR+1:
		case OP_LD_MRR+2:
		case OP_LD_MRR+3:
			osz = tsz[**cp - OP_LD_MRR];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			if (i1 + osz > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			memcpy(&(x->mem[i1]), x->reg[r2], osz);
			break;
		case OP_ADD:
		case OP_ADD+1:
		case OP_ADD+2:
		case OP_ADD+3:
			osz = tsz[**cp - OP_ADD];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);
			i1 += i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_FADD:
		case OP_FADD+1:
		case OP_FADD+2:
		case OP_FADD+3:
			i1 = **cp - OP_FADD;
			osz = tsz[i1];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			read_float(x->reg[r1], (int)i1, &mn1, &ep1);
			read_float(x->reg[r2], (int)i1, &mn2, &ep2);

			if (ep1 < ep2)
			{
				mn1 >>= ep2 - ep1;
				ep1 = ep2;
			}
			else if (ep2 < ep1)
			{
				mn2 >>= ep1 - ep2;
				ep2 = ep1;
			}

			mn1 += mn2;
			write_float(x->reg[r1], mn1, ep1, (int)i1);
			read_float(x->reg[r1], (int)i1, &mn1, &ep1);

			break;
		case OP_SUB:
		case OP_SUB+1:
		case OP_SUB+2:
		case OP_SUB+3:
			osz = tsz[**cp - OP_SUB];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);
			i1 -= i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_FSUB:
		case OP_FSUB+1:
		case OP_FSUB+2:
		case OP_FSUB+3:
			i1 = **cp - OP_FSUB;
			osz = tsz[i1];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			read_float(x->reg[r1], (int)i1, &mn1, &ep1);
			read_float(x->reg[r2], (int)i1, &mn2, &ep2);

			if (ep1 < ep2)
			{
				mn1 >>= ep2 - ep1;
				ep1 = ep2;
			}
			else if (ep2 < ep1)
			{
				mn2 >>= ep1 - ep2;
				ep2 = ep1;
			}

			mn1 -= mn2;
			write_float(x->reg[r1], mn1, ep1, (int)i1);

			break;
		case OP_MULT:
		case OP_MULT+1:
		case OP_MULT+2:
		case OP_MULT+3:
			osz = tsz[**cp - OP_MULT];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);
			i1 *= i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_FMULT:
		case OP_FMULT+1:
		case OP_FMULT+2:
		case OP_FMULT+3:
			i1 = **cp - OP_FMULT;
			osz = tsz[**cp - OP_FMULT];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			read_float(x->reg[r1], (int)i1, &mn1, &ep1);
			read_float(x->reg[r2], (int)i1, &mn2, &ep2);

			mn1 >>= cull[i1];
			mn2 >>= cull[i1];
			mn1 *= mn2;
			ep1 += cull[i1]*2 + ep2 - mantissa[i1];
			write_float(x->reg[r1], mn1, ep1, (int)i1);

			break;
		case OP_MULTS:
		case OP_MULTS+1:
		case OP_MULTS+2:
		case OP_MULTS+3:
			ep1 = **cp - OP_MULTS;
			osz = tsz[ep1];
			im = imax[ep1];

			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);

			if (i1 > msb[ep1])
				s1 = -(signed _int64)(im - i1 + 1);
			else
				s1 = i1;

			if (i2 > msb[ep1])
				s2 = -(signed _int64)(im - i2 + 1);
			else
				s2 = i2;

			s1 *= s2;

			if (s1 < 0)
				i1 = im + s1 + 1;
			else
				i1 = s1;

			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}

			break;
		case OP_DIV:
		case OP_DIV+1:
		case OP_DIV+2:
		case OP_DIV+3:
			osz = tsz[**cp - OP_DIV];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);
			if (!i2)
			{
				xc_set_errno(XC_ERR_DIV_BY_ZERO);
				return 0;
			}
			i1 /= i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_FDIV:
		case OP_FDIV+1:
		case OP_FDIV+2:
		case OP_FDIV+3:
			i1 = **cp - OP_FDIV;
			osz = tsz[**cp - OP_FDIV];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			read_float(x->reg[r1], (int)i1, &mn1, &ep1);
			read_float(x->reg[r2], (int)i1, &mn2, &ep2);

			if (!mn2)
				ep1 = fullexp[i1];
			else
			{
				mn1 = (_int64)((double)mn1 / (double)mn2 * divbump1[i1]);
				ep1 -= ep2 - mantissa[i1] + divbump2[i1];
			}

			write_float(x->reg[r1], mn1, ep1, (int)i1);

			break;
		case OP_DIVS:
		case OP_DIVS+1:
		case OP_DIVS+2:
		case OP_DIVS+3:
			ep1 = **cp - OP_DIVS;
			osz = tsz[ep1];
			im = imax[ep1];

			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);

			if (!i2)
			{
				xc_set_errno(XC_ERR_DIV_BY_ZERO);
				return 0;
			}

			if (i1 > msb[ep1])
				s1 = -(signed _int64)(im - i1 + 1);
			else
				s1 = i1;

			if (i2 > msb[ep1])
				s2 = -(signed _int64)(im - i2 + 1);
			else
				s2 = i2;

			s1 /= s2;

			if (s1 < 0)
				i1 = im + s1 + 1;
			else
				i1 = s1;

			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}

			break;
		case OP_MOD:
		case OP_MOD+1:
		case OP_MOD+2:
		case OP_MOD+3:
			osz = tsz[**cp - OP_MOD];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);
			if (!i2)
			{
				xc_set_errno(XC_ERR_DIV_BY_ZERO);
				return 0;
			}
			i1 %= i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_MODS:
		case OP_MODS+1:
		case OP_MODS+2:
		case OP_MODS+3:
			ep1 = **cp - OP_DIVS;
			osz = tsz[ep1];
			im = imax[ep1];

			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);

			if (!i2)
			{
				xc_set_errno(XC_ERR_DIV_BY_ZERO);
				return 0;
			}

			if (i1 > msb[ep1])
				s1 = -(signed _int64)(im - i1 + 1);
			else
				s1 = i1;

			if (i2 > msb[ep1])
				s2 = -(signed _int64)(im - i2 + 1);
			else
				s2 = i2;

			s1 %= s2;

			if (s1 < 0)
				i1 = im + s1 + 1;
			else
				i1 = s1;

			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}

			break;
		case OP_SL:
		case OP_SL+1:
		case OP_SL+2:
		case OP_SL+3:
			osz = tsz[**cp - OP_SL];
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i1 <<= 1;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_SLN:
		case OP_SLN+1:
		case OP_SLN+2:
		case OP_SLN+3:
			osz = tsz[**cp - OP_SLN];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = x->reg[r2][0];
			i1 <<= i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_SR:
		case OP_SR+1:
		case OP_SR+2:
		case OP_SR+3:
			osz = tsz[**cp - OP_SR];
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i1 >>= 1;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_SRN:
		case OP_SRN+1:
		case OP_SRN+2:
		case OP_SRN+3:
			osz = tsz[**cp - OP_SRN];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = x->reg[r2][0];
			i1 >>= i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_NEG:
		case OP_NEG+1:
		case OP_NEG+2:
		case OP_NEG+3:
			osz = tsz[**cp - OP_NEG];
			i3 = makeneg[**cp - OP_NEG];
			im = imax[**cp - OP_NEG];

			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);

			i1 = im - i1 + 1;

			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_FNEG:
		case OP_FNEG+1:
		case OP_FNEG+2:
		case OP_FNEG+3:
			i1 = **cp - OP_FNEG;
			osz = tsz[i1];
			(*cp)++;
			r1 = **cp & 15;

			read_float(x->reg[r1], (int)i1, &mn1, &ep1);
			mn1 = -mn1;
			write_float(x->reg[r1], mn1, ep1, (int)i1);

			break;
		case OP_AND:
		case OP_AND+1:
		case OP_AND+2:
		case OP_AND+3:
			osz = tsz[**cp - OP_AND];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);
			i1 &= i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_OR:
		case OP_OR+1:
		case OP_OR+2:
		case OP_OR+3:
			osz = tsz[**cp - OP_OR];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);
			i1 |= i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_XOR:
		case OP_XOR+1:
		case OP_XOR+2:
		case OP_XOR+3:
			osz = tsz[**cp - OP_XOR];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);
			i1 ^= i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_NOT:
		case OP_NOT+1:
		case OP_NOT+2:
		case OP_NOT+3:
			osz = tsz[**cp - OP_NOT];
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i1 = ~i1;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_BAND:
		case OP_BAND+1:
		case OP_BAND+2:
		case OP_BAND+3:
			osz = tsz[**cp - OP_BAND];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);
			i1 = i1 && i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_BOR:
		case OP_BOR+1:
		case OP_BOR+2:
		case OP_BOR+3:
			osz = tsz[**cp - OP_BOR];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += *sp2++ << (8 * k);
			i1 = i1 || i2;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_BNOT:
		case OP_BNOT+1:
		case OP_BNOT+2:
		case OP_BNOT+3:
			osz = tsz[**cp - OP_BNOT];
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i1 = !i1;

			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_CP:
		case OP_CP+1:
		case OP_CP+2:
		case OP_CP+3:
			osz = tsz[**cp - OP_CP];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += (_int64)*sp2++ << (8 * k);
			if (i1 == i2)
				x->z = 1;
			else
				x->z = 0;
			if (i1 < i2)
				x->c = 1;
			else
				x->c = 0;
			break;
		case OP_SCP:
		case OP_SCP+1:
		case OP_SCP+2:
		case OP_SCP+3:
			ep1 = **cp - OP_SCP;
			osz = tsz[ep1];
			im = imax[ep1];

			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < osz; ++k)
				i2 += (_int64)*sp2++ << (8 * k);

			if (i1 > msb[ep1])
				s1 = -(signed _int64)(im - i1 + 1);
			else
				s1 = i1;

			if (i2 > msb[ep1])
				s2 = -(signed _int64)(im - i2 + 1);
			else
				s2 = i2;

			if (s1 == s2)
				x->z = 1;
			else
				x->z = 0;
			if (s1 < s2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_FCP:
		case OP_FCP+1:
		case OP_FCP+2:
		case OP_FCP+3:
			i1 = **cp - OP_FCP;
			osz = tsz[**cp - OP_FCP];
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			read_float(x->reg[r1], (int)i1, &mn1, &ep1);
			read_float(x->reg[r2], (int)i1, &mn2, &ep2);

			db1 = mn1 * pow(2, ep1 - mantissa[i1]);
			db2 = mn2 * pow(2, ep2 - mantissa[i1]);

			x->z = (db1 == db2);
			x->c = (db1 < db2);

			break;
		case OP_TEST:
		case OP_TEST+1:
		case OP_TEST+2:
		case OP_TEST+3:
			osz = tsz[**cp - OP_TEST];
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			x->z = !i1;
			x->c = 0;
			break;
		case OP_STEST:
		case OP_STEST+1:
		case OP_STEST+2:
		case OP_STEST+3:
			ep1 = **cp - OP_DIVS;
			osz = tsz[ep1];
			im = imax[ep1];
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			if (i1 > msb[ep1])
				s1 = -(signed _int64)(im - i1 + 1);
			else
				s1 = i1;
			x->z = !s1;
			x->c = s1 < 0;
			break;
		case OP_JP:
			i1 = 0;
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*++*cp << (k * 8);
			*cp = code + i1 - 1;
			break;
		case OP_JP_Z:
		case OP_JP_NZ:
		case OP_JP_C:
		case OP_JP_NC:
			switch (**cp) {
				case OP_JP_Z:
					ch = x->z;
					break;
				case OP_JP_NZ:
					ch = !x->z;
					break;
				case OP_JP_C:
					ch = x->c;
					break;
				case OP_JP_NC:
					ch = !x->c;
				}
			i1 = 0;
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*++*cp << (k * 8);
			if (ch)
				*cp = code + i1 - 1;
			break;
		case OP_JP_R:
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (k * 8);
			*cp = code + i1 - 1;
			break;
		case OP_JP_R_Z:
		case OP_JP_R_NZ:
		case OP_JP_R_C:
		case OP_JP_R_NC:
			switch (**cp)
			{
				case OP_JP_R_Z:
					ch = x->z;
					break;
				case OP_JP_R_NZ:
					ch = !x->z;
					break;
				case OP_JP_R_C:
					ch = x->c;
					break;
				case OP_JP_R_NC:
					ch = !x->c;
			}
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (k * 8);
			if (ch)
				*cp = code + i1 - 1;
			break;
		case OP_INC:
		case OP_INC+1:
		case OP_INC+2:
		case OP_INC+3:
			osz = tsz[**cp - OP_INC];
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			++i1;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_DEC:
		case OP_DEC+1:
		case OP_DEC+2:
		case OP_DEC+3:
			osz = tsz[**cp - OP_DEC];
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			--i1;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_FINC:
		case OP_FINC+1:
		case OP_FINC+2:
		case OP_FINC+3:
			i1 = **cp - OP_FINC;
			osz = tsz[i1];
			(*cp)++;
			r1 = **cp & 15;

			read_float(x->reg[r1], (int)i1, &mn1, &ep1);
			mn1 += (_int64)1 << (mantissa[i1] - ep1);
			write_float(x->reg[r1], mn1, ep1, (int)i1);

			break;
		case OP_FDEC:
		case OP_FDEC+1:
		case OP_FDEC+2:
		case OP_FDEC+3:
			i1 = **cp - OP_FDEC;
			osz = tsz[i1];
			(*cp)++;
			r1 = **cp & 15;

			read_float(x->reg[r1], (int)i1, &mn1, &ep1);
			mn1 -= (_int64)1 << (mantissa[i1] - ep1);
			write_float(x->reg[r1], mn1, ep1, (int)i1);

			break;
		case OP_PUSH:
		case OP_PUSH+1:
		case OP_PUSH+2:
		case OP_PUSH+3:
			SPREAD;
			osz = tsz[**cp - OP_PUSH];
			(*cp)++;
			r1 = **cp & 15;
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			memcpy(s->n, x->reg[r1], osz);
			break;
		case OP_POP:
		case OP_POP+1:
		case OP_POP+2:
		case OP_POP+3:
			if (!x->stack)
			{
				xc_set_errno(XC_ERR_STACK_UNDERFLOW);
				return 0;
			}
			osz = tsz[**cp - OP_POP];
			(*cp)++;
			r1 = **cp & 15;
			memcpy(x->reg[r1], x->stack->n, osz);
			s = x->stack;
			x->stack = x->stack->next;
			free(s);
			break;
		case OP_CALL_C:
			i1 = 0;
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*++*cp << (k * 8);
			p = (pstack *)calloc(1, sizeof(pstack));
			p->next = x->cstack;
			x->cstack = p;
			x->cstack->ptr = *cp + 1;
			*cp = code + i1 - 1;
			break;
		case OP_CALL_R:
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (k * 8);
			p = (pstack *)calloc(1, sizeof(pstack));
			p->next = x->cstack;
			x->cstack = p;
			x->cstack->ptr = *cp + 1;
			*cp = code + i1 - 1;
			break;
		case OP_ARG_R:
		case OP_ARG_R+1:
		case OP_ARG_R+2:
		case OP_ARG_R+3:
			r2 = **cp - OP_ARG_R;
			osz = tsz[r2];
			im = imax[r2];

			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*sp1++ << (k * 8);

			x->arg = (_int64 *)realloc(x->arg, sizeof(_int64) * ++x->narg);
			x->arg[x->narg-1] = i1;
			break;
		case OP_ARG_C:
		case OP_ARG_C+1:
		case OP_ARG_C+2:
		case OP_ARG_C+3:
			r2 = **cp - OP_ARG_C;
			osz = tsz[r2];
			i3 = makeneg[r2];
			im = imax[r2];

			i1 = 0;
			for (k = 0; k < osz; ++k)
				i1 += (_int64)*++*cp << (k * 8);

			x->arg = (_int64 *)realloc(x->arg, sizeof(_int64) * ++x->narg);
			x->arg[x->narg-1] = i1;
			break;
		case OP_HCALL_C:
			i1 = 0;
			for (k = 0; k < 2; ++k)
				i1 += (_int64)*++*cp << (k * 8);

			if (i1 >= n_hostfunc)
			{
				xc_set_errno(XC_ERR_HOSTSYM_SEGFAULT);
				return 0;
			}

			arg = calloc(narg = x->narg, sizeof(_int64));
			memcpy(arg, x->arg, sizeof(_int64) * narg);
			free(x->arg);
			x->arg = malloc(0);
			x->narg = 0;
			i1 = hostfunc[i1](x, narg, arg);
			free(arg);
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			sp1 = s->n;
			for (k = 0; k < S_REG; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}

			break;
		case OP_HCALL_R:
			(*cp)++;
			r1 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 2; ++k)
				i1 += (_int64)*sp1++ << (k * 8);
			if (i1 >= n_hostfunc)
			{
				xc_set_errno(XC_ERR_HOSTSYM_SEGFAULT);
				return 0;
			}
			arg = calloc(narg = x->narg, sizeof(_int64));
			memcpy(arg, x->arg, sizeof(_int64) * narg);
			free(x->arg);
			x->arg = malloc(0);
			x->narg = 0;
			i1 = hostfunc[i1](x, narg, arg);
			free(arg);
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			sp1 = s->n;
			for (k = 0; k < S_REG; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}
			break;
		case OP_RET:
			if (!x->cstack)
			{
				//unload_xinst(x);
				xc_set_errno(XC_ERR_NONE);
				return 0;
			}
			*cp = x->cstack->ptr - 1;
			p = x->cstack;
			x->cstack = p->next;
			free(p);
			break;
		case OP_RETV:
		case OP_RETV+1:
		case OP_RETV+2:
		case OP_RETV+3:
			if (!x->cstack)
			{
				//unload_xinst(x);
				xc_set_errno(XC_ERR_NONE);
				return 0;
			}
			osz = tsz[**cp - OP_RETV];
			(*cp)++;
			r1 = **cp & 15;
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			memcpy(s->n, x->reg[r1], osz);

			*cp = x->cstack->ptr - 1;
			p = x->cstack;
			x->cstack = p->next;
			free(p);
			break;
		case OP_CPY:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			(*cp)++;
			r3 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp1 = x->reg[r2];
			for (k = 0; k < 4; ++k)
				i2 += *sp1++ << (8 * k);
			i3 = 0;
			sp1 = x->reg[r3];
			for (k = 0; k < 4; ++k)
				i3 += *sp1++ << (8 * k);

			if (i1 + i3 > x->memsz || i2 + i3 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}

			memcpy(&(x->mem[i1]), &(x->mem[i2]), (int)i3);
			break;
		case OP_SCPY:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp1 = x->reg[r2];
			for (k = 0; k < 4; ++k)
				i2 += *sp1++ << (8 * k);

			if (i1 >= x->memsz || i2 >= x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}

			strcpy(&(x->mem[i1]), &(x->mem[i2]));
			break;
		case OP_DCPY:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp1 = x->reg[r2];
			for (k = 0; k < 4; ++k)
				i2 += *sp1++ << (8 * k);

			if (i1 >= x->memsz || i2 >= x->parent->n_data)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}

			memcpy(&(x->mem[i1]), x->parent->data[i2].dat, x->parent->data[i2].sz);
			break;
		case OP_PMODE:
		case OP_IMODE:
		case OP_RMODE:
		case OP_OMODE:
//				printf("Mode change %u\n", (unsigned)**cp);
			x->mode = **cp;
			s8 = (stack8 *)calloc(1, sizeof(stack8));
			s8->next = x->cmode;
			x->cmode = s8;
			s8->n = **cp;
			break;
		case OP_EMODE:
//				puts("Mode end");
			if (x->cmode)
			{
				s8 = x->cmode;
				x->cmode = s8->next;
				free(s8);
				if (x->cmode)
					x->mode = x->cmode->n;
				else
					x->mode = OP_PMODE;
			}
			else
				x->mode = OP_PMODE;
			break;
		case OP_SYNC:
			if (x->serv)
			{
				++*cp;
				return -1;
			}
			break;
		case OP_RSEND:
			if (x->serv)
			{
				++*cp;
				if (x->serv->mode != XN_ALTER)
					x->serv->lmode = x->serv->mode;
				x->serv->mode = XN_TRANS;
				x->tmode = TRANS_RECV;
				x->transp = x->reg[**cp & 15];
				x->transz = 8;
				++*cp;

				do_recv(x);

				x->serv->mode = x->serv->lmode;
				x->tmode = TRANS_NONE;

				return -1;
			}
			else
				++*cp;

			break;
		case OP_MSEND:
			if (x->serv)
			{
				++*cp;
				if (x->serv->mode != XN_ALTER)
					x->serv->lmode = x->serv->mode;
				x->serv->mode = XN_TRANS;
				x->tmode = TRANS_RECV;
				r1 = **cp >> 4;
				r2 = **cp & 15;
				i1 = 0;
				sp1 = x->reg[r1];
				for (k = 0; k < 4; ++k)
					i1 += (_int64)*sp1++ << (8 * k);
				i2 = 0;
				sp2 = x->reg[r2];
				for (k = 0; k < 4; ++k)
					i2 += *sp2++ << (8 * k);
				x->transp = &x->mem[i1];
				x->transz = (int)i2;
				++*cp;

				do_recv(x);

				x->serv->mode = x->serv->lmode;
				x->tmode = TRANS_NONE;

				return -1;
			}
			else
				++*cp;

			break;
		case OP_MBASE:
			i1 = 0;
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*++*cp << (8 * k);
			if (i1 > x->memsz)
			{
				xc_set_errno(XC_ERR_MODULE_MEMORY);
				return 0;
			}

			if (x->mb_base)
			{
				xc_set_errno(XC_ERR_MODULE_MEMORY);
				return 0;
			}

			mb = (memblock *)calloc(1, sizeof(memblock));
			mb->min = 0;
			mb->max = (int)i1;

			x->mb_base = x->cur_mb = mb;

			break;
		case OP_MALLOC:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			i1 = 0;
			sp1 = x->reg[r2];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (8 * k);

			if ((i2 = __xc_malloc(x, (int)i1)) < 0)
			{
				puts("This here!");
				return 0;
			}

			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
			{
				*sp1++ = (char)(i2 & 255);
				i2 >>= 8;
			}

			break;
		case OP_FREE:
			(*cp)++;
			r1 = **cp & 15;

			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (8 * k);

			__xc_free(x, (int)i1);
			if (xc_get_errno())
			{
				puts("Free crash!");
				return 0;
			}

			break;
		case OP_LALLOC:
			i1 = 0;
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*++*cp << (8 * k);

			if ((i2 = __xc_malloc(x, (int)i1)) < 0)
				return 0;

			ln = (lint *)calloc(1, sizeof(lint));
			ln->next = x->local;
			if (x->local)
				x->local->prev = ln;
			x->local = ln;
			ln->n = (int)i2;

			break;
		case OP_LFREE:
			if (!x->local)
			{
				xc_set_errno(XC_ERR_STACK_UNDERFLOW);
				return 0;
			}

			__xc_free(x, x->local->n);
			if (xc_get_errno())
			{
				puts("Lfree crashed!");
				return 0;
			}

			if (x->local->next)
				x->local->next->prev = NULL;
			ln = x->local;
			x->local = x->local->next;
			free(ln);

			break;
		case OP_LADDR:
			i1 = 0;
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*++*cp << (8 * k);

			if (!x->local)
			{
				xc_set_errno(XC_ERR_STACK_UNDERFLOW);
				return 0;
			}

			i2 = x->local->n + i1;
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			sp1 = s->n;
			for (k = 0; k < 4; ++k)
			{
				*sp1++ = (char)(i2 & 255);
				i2 >>= 8;
			}

			break;
		case OP_CAST:
			{
				char t1, t2, sn1, sn2, sz1, sz2;

				if (!x->stack)
				{
					xc_set_errno(XC_ERR_STACK_UNDERFLOW);
					return 0;
				}

				i8_1 = *++*cp;
				t1 = (char)(i8_1 & 3);
				i8_1 >>= 2;
				sz1 = (char)(i8_1 & 3);
				sn1 = tsz[i8_1 & 3];
				i8_1 >>= 2;
				t2 = (char)(i8_1 & 3);
				i8_1 >>= 2;
				sz2 = (char)(i8_1 & 3);
				sn2 = tsz[i8_1 & 3];

//				printf("%d/%d vs. %d/%d\n", t1, sn1, t2, sn2);
				if (t1 == _C_FLOAT && t2 == _C_FLOAT)
				{
					read_float(x->stack->n, sz1, &mn1, &ep1);
					if (sn1 > sn2)
						mn1 >>= mantissa[sz1] - mantissa[sz2];
					else
						mn1 <<= mantissa[sz2] - mantissa[sz1];
					write_float(x->stack->n, mn1, ep1, sz2);
					read_float(x->stack->n, sz2, &mn1, &ep1);
				}
				else if (t1 == _C_FLOAT)
				{
					read_float(x->stack->n, sz1, &mn1, &ep1);
					write_int(x->stack->n, sz2, mn1 >> (mantissa[sz1] - ep1));
				}
				else if (t2 == _C_FLOAT)
				{
					read_int(x->stack->n, sz1, &mn1);
					write_float(x->stack->n, mn1, mantissa[sz2], sz2);
					read_float(x->stack->n, sz2, &mn1, &ep1);
				}
				else
				{
					if (t1 == _C_SIGNED)
					{
						if (sn2 > sn1)
						{
							if (x->stack->n[sn1-1] & 128)
							{
								i64_1 = 0;
								sp1 = x->stack->n;
								for (k = 0; k < sn1; ++k)
									i64_1 += (_int64)*sp1++ << (8 * k);
								i64_1 += imax[sz2] - imax[sz1];
								sp1 = x->stack->n;
								for (k = 0; k < sn2; ++k)
								{
									*sp1++ = (char)(i64_1 & 255);
									i64_1 >>= 8;
								}
							}
							else
								memset(&x->stack->n[sn1],
									0, sn2 - sn1);
						}
					}
					else if (sn2 > sn1)
						memset(&x->stack->n[sn1],
							0, sn2 - sn1);
				}

				*cp += 3;
			}

			break;
		case OP_GBF:
			(*cp)++;
			r1 = **cp & 15;

			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += *sp1++ << (8 * k);
			ui1 = *++*cp;
			ui2 = *++*cp;

			i1 >>= ui1;
			i1 &= ((_int64)1 << ui2) - 1;

			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}

			break;
		case OP_SBF:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += *sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < 4; ++k)
				i2 += *sp2++ << (8 * k);
			ui1 = *++*cp;
			ui2 = *++*cp;

			i2 &= ((_int64)1 << ui2) - 1;
			i2 <<= ui1;
			for (ku = 0; ku < ui2; ++ku)
				if (i1 & (1 << (ui1 + ku)))
					i1 ^= 1 << (ui1 + ku);
			i1 |= i2;

			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}

			break;
		case OP_SGBF:
			(*cp)++;
			r1 = **cp & 15;

			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += *sp1++ << (8 * k);
			ui1 = *++*cp;
			ui2 = *++*cp;

			i1 >>= ui1;
			i1 &= ((_int64)1 << ui2) - 1;

			if (i1 & (1 << (ui2-1)))
				i1 = 0xFFFFFFFF - ((0xFFFFFFFF & ((1 << ui2) - 1)) - i1);

			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}

			break;
		case OP_SSBF:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += *sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < 4; ++k)
				i2 += *sp2++ << (8 * k);
			ui1 = *++*cp;
			ui2 = *++*cp;

			if (i2 & 0x80000000)
				i2 = (0xFFFFFF & ((1 << ui2) - 1)) - (0x80000000 - i2);

			i2 &= ((_int64)1 << ui2) - 1;
			i2 <<= ui1;
			for (ku = 0; ku < ui2; ++ku)
				if (i1 & (1 << (ui1 + ku)))
					i1 ^= 1 << (ui1 + ku);
			i1 |= i2;

			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
			{
				*sp1++ = (char)(i1 & 255);
				i1 >>= 8;
			}

			break;
		case OP_ZERO:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			i1 = 0;
			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
				i1 += (_int64)*sp1++ << (8 * k);
			i2 = 0;
			sp2 = x->reg[r2];
			for (k = 0; k < 4; ++k)
				i2 += *sp2++ << (8 * k);

			if ((i1 + i2) < 0 || (i1 + i2) >= x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}

			memset(&x->mem[i1], 0, (int)i2);
			break;
		case OP_SLEN:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			i1 = 0;
			sp1 = x->reg[r2];
			for (k = 0; k < 4; ++k)
				i1 += *sp1++ << (8 * k);

			if (!i1)
			{
				memset(x->reg[r1], 0, 4);
				break;
			}

			i2 = 1;
			for (; ; ++i1)
			{
				if (i1 >= x->memsz)
				{
					xc_set_errno(XC_ERR_SEGFAULT);
					return 0;
				}
				if (!x->mem[i1])
					break;
				++i2;
			}

			sp1 = x->reg[r1];
			for (k = 0; k < 4; ++k)
			{
				*sp1++ = (char)(i2 & 255);
				i2 >>= 8;
			}

			break;
		case OP_PING:
			if (x->serv)
				xnet_sendc(x->serv, (char)OP_PONG);

			break;
#else
		case OP_LD_RR:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			memcpy(x->reg[r1], x->reg[r2], 1);

			break;
		case OP_LD_RR+1:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			memcpy(x->reg[r1], x->reg[r2], 2);

			break;
		case OP_LD_RR+2:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			memcpy(x->reg[r1], x->reg[r2], 4);

			break;
		case OP_LD_RR+3:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			memcpy(x->reg[r1], x->reg[r2], 8);

			break;
		case OP_LD_RC:
			(*cp)++;
			r1 = **cp & 15;

			x->reg[r1][0] = *++*cp;

			break;
		case OP_LD_RC+1:
			(*cp)++;
			r1 = **cp & 15;

			x->reg[r1][0] = *++*cp;
			x->reg[r1][1] = *++*cp;

			break;
		case OP_LD_RC+2:
			(*cp)++;
			r1 = **cp & 15;

			x->reg[r1][0] = *++*cp;
			x->reg[r1][1] = *++*cp;
			x->reg[r1][2] = *++*cp;
			x->reg[r1][3] = *++*cp;

			break;
		case OP_LD_RC+3:
			(*cp)++;
			r1 = **cp & 15;

			memcpy(x->reg[r1], ++*cp, 8);
			*cp += 7;

			break;
		case OP_LD_RMC:
			(*cp)++;
			r1 = **cp & 15;
			i32_1 = *(_uint32 *)++*cp;
			*cp += 3;
			if (i32_1 + 1 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->reg[r1][0] = x->mem[i32_1];

			break;
		case OP_LD_RMC+1:
			(*cp)++;
			r1 = **cp & 15;
			i32_1 = *(_uint32 *)++*cp;
			*cp += 3;
			if (i32_1 + 2 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->reg[r1][0] = x->mem[i32_1];
			x->reg[r1][0] = x->mem[++i32_1];

			break;
		case OP_LD_RMC+2:
			(*cp)++;
			r1 = **cp & 15;
			i32_1 = *(_uint32 *)++*cp;
			*cp += 3;
			if (i32_1 + 4 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->reg[r1][0] = x->mem[i32_1];
			x->reg[r1][1] = x->mem[++i32_1];
			x->reg[r1][2] = x->mem[++i32_1];
			x->reg[r1][3] = x->mem[++i32_1];

			break;
		case OP_LD_RMC+3:
			(*cp)++;
			r1 = **cp & 15;
			i32_1 = *(_uint32 *)++*cp;
			*cp += 3;
			if (i32_1 + 8 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}

			memcpy(x->reg[r1], &x->mem[i32_1], 8);

			break;
		case OP_LD_RMR:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i32_1 = *(_uint32 *)x->reg[r2];
			if (i32_1 + 1 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->reg[r1][0] = x->mem[i32_1];

			break;
		case OP_LD_RMR+1:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i32_1 = *(_uint32 *)x->reg[r2];
			if (i32_1 + 2 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->reg[r1][0] = x->mem[i32_1];
			x->reg[r1][1] = x->mem[++i32_1];

			break;
		case OP_LD_RMR+2:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i32_1 = *(_uint32 *)x->reg[r2];
			if (i32_1 + 4 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->reg[r1][0] = x->mem[i32_1];
			x->reg[r1][1] = x->mem[++i32_1];
			x->reg[r1][2] = x->mem[++i32_1];
			x->reg[r1][3] = x->mem[++i32_1];

			break;
		case OP_LD_RMR+3:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i32_1 = *(_uint32 *)x->reg[r2];
			if (i32_1 + 8 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			memcpy(x->reg[r1], &x->mem[i32_1], 8);

			break;
		case OP_LD_MCR:
			i32_1 = *(_uint32 *)++*cp;
			r2 = *(*cp += 4) & 15;
			if (i32_1 + 1 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->mem[i32_1] = x->reg[r2][0];

			break;
		case OP_LD_MCR+1:
			i32_1 = *(_uint32 *)++*cp;;
			r2 = *(*cp += 4) & 15;
			if (i32_1 + 2 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->mem[i32_1] = x->reg[r2][0];
			x->mem[++i32_1] = x->reg[r2][1];

			break;
		case OP_LD_MCR+2:
			i32_1 = *(_uint32 *)++*cp;
			r2 = *(*cp += 4) & 15;
			if (i32_1 + 4 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->mem[i32_1] = x->reg[r2][0];
			x->mem[++i32_1] = x->reg[r2][1];
			x->mem[++i32_1] = x->reg[r2][2];
			x->mem[++i32_1] = x->reg[r2][3];

			break;
		case OP_LD_MCR+3:
			i32_1 = *(_uint32 *)++*cp;;
			r2 = *(*cp += 4) & 15;
			if (i32_1 + 8 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			memcpy(&x->mem[i32_1], x->reg[r2], 8);

			break;
		case OP_LD_MRR:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i32_1 = *(_uint32 *)x->reg[r1];
			if (i32_1 + 1 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->mem[i32_1] = x->reg[r2][0];

			break;
		case OP_LD_MRR+1:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i32_1 = *(_uint32 *)x->reg[r1];
			if (i32_1 + 2 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->mem[i32_1] = x->reg[r2][0];
			x->mem[++i32_1] = x->reg[r2][1];

			break;
		case OP_LD_MRR+2:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i32_1 = *(_uint32 *)x->reg[r1];
			if (i32_1 + 4 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			x->mem[i32_1] = x->reg[r2][0];
			x->mem[++i32_1] = x->reg[r2][1];
			x->mem[++i32_1] = x->reg[r2][2];
			x->mem[++i32_1] = x->reg[r2][3];

			break;
		case OP_LD_MRR+3:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;
			i32_1 = *(_uint32 *)x->reg[r1];
			if (i32_1 + 8 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}
			memcpy(&x->mem[i32_1], x->reg[r2], 8);

			break;
		case OP_ADD:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] + *x->reg[**cp & 15];

			break;
		case OP_ADD+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] +
				*(_uint16 *)x->reg[**cp & 15];

			break;
		case OP_ADD+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] +
				*(_uint32 *)x->reg[**cp & 15];

			break;
		case OP_ADD+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] +
				*(_uint64 *)x->reg[**cp & 15];

			break;
		case OP_FADD+2:
			(*cp)++;
			*(_float32 *)x->reg[**cp >> 4] = *(_float32 *)x->reg[**cp >> 4] +
				*(_float32 *)x->reg[**cp & 15];

			break;
		case OP_FADD+3:
			(*cp)++;
			*(_float64 *)x->reg[**cp >> 4] = *(_float64 *)x->reg[**cp >> 4] +
				*(_float64 *)x->reg[**cp & 15];

			break;
		case OP_SUB:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] - *x->reg[**cp & 15];

			break;
		case OP_SUB+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] -
				*(_uint16 *)x->reg[**cp & 15];

			break;
		case OP_SUB+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] -
				*(_uint32 *)x->reg[**cp & 15];

			break;
		case OP_SUB+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] -
				*(_uint64 *)x->reg[**cp & 15];

			break;
		case OP_FSUB+2:
			(*cp)++;
			*(_float32 *)x->reg[**cp >> 4] = *(_float32 *)x->reg[**cp >> 4] -
				*(_float32 *)x->reg[**cp & 15];

			break;
		case OP_FSUB+3:
			(*cp)++;
			*(_float64 *)x->reg[**cp >> 4] = *(_float64 *)x->reg[**cp >> 4] -
				*(_float64 *)x->reg[**cp & 15];

			break;
		case OP_MULT:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] * *x->reg[**cp & 15];

			break;
		case OP_MULT+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] *
				*(_uint16 *)x->reg[**cp & 15];

			break;
		case OP_MULT+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] *
				*(_uint32 *)x->reg[**cp & 15];

			break;
		case OP_MULT+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] *
				*(_uint64 *)x->reg[**cp & 15];

			break;
		case OP_FMULT+2:
			(*cp)++;
			*(_float32 *)x->reg[**cp >> 4] = *(_float32 *)x->reg[**cp >> 4] *
				*(_float32 *)x->reg[**cp & 15];

			break;
		case OP_FMULT+3:
			(*cp)++;
			*(_float64 *)x->reg[**cp >> 4] = *(_float64 *)x->reg[**cp >> 4] *
				*(_float64 *)x->reg[**cp & 15];

			break;
		case OP_MULTS:
			(*cp)++;
			*(char *)x->reg[**cp >> 4] = *(char *)x->reg[**cp >> 4] *
				*(char *)x->reg[**cp & 15];

			break;
		case OP_MULTS+1:
			(*cp)++;
			*(_int16 *)x->reg[**cp >> 4] = *(_int16 *)x->reg[**cp >> 4] *
				*(_int16 *)x->reg[**cp & 15];

			break;
		case OP_MULTS+2:
			(*cp)++;
			*(_int32 *)x->reg[**cp >> 4] = *(_int32 *)x->reg[**cp >> 4] *
				*(_int32 *)x->reg[**cp & 15];

			break;
		case OP_MULTS+3:
			(*cp)++;
			*(_int64 *)x->reg[**cp >> 4] = *(_int64 *)x->reg[**cp >> 4] *
				*(_int64 *)x->reg[**cp & 15];

			break;
		case OP_DIV:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] / *x->reg[**cp & 15];

			break;
		case OP_DIV+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] /
				*(_uint16 *)x->reg[**cp & 15];

			break;
		case OP_DIV+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] /
				*(_uint32 *)x->reg[**cp & 15];

			break;
		case OP_DIV+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] /
				*(_uint64 *)x->reg[**cp & 15];

			break;
		case OP_FDIV+2:
			(*cp)++;
			*(_float32 *)x->reg[**cp >> 4] = *(_float32 *)x->reg[**cp >> 4] /
				*(_float32 *)x->reg[**cp & 15];

			break;
		case OP_FDIV+3:
			(*cp)++;
			*(_float64 *)x->reg[**cp >> 4] = *(_float64 *)x->reg[**cp >> 4] /
				*(_float64 *)x->reg[**cp & 15];

			break;
		case OP_DIVS:
			(*cp)++;
			*(char *)x->reg[**cp >> 4] = *(char *)x->reg[**cp >> 4] /
				*(char *)x->reg[**cp & 15];

			break;
		case OP_DIVS+1:
			(*cp)++;
			*(_int16 *)x->reg[**cp >> 4] = *(_int16 *)x->reg[**cp >> 4] /
				*(_int16 *)x->reg[**cp & 15];

			break;
		case OP_DIVS+2:
			(*cp)++;
			*(_int32 *)x->reg[**cp >> 4] = *(_int32 *)x->reg[**cp >> 4] /
				*(_int32 *)x->reg[**cp & 15];

			break;
		case OP_DIVS+3:
			(*cp)++;
			*(_int64 *)x->reg[**cp >> 4] = *(_int64 *)x->reg[**cp >> 4] /
				*(_int64 *)x->reg[**cp & 15];

			break;
		case OP_MOD:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] / *x->reg[**cp & 15];

			break;
		case OP_MOD+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] %
				*(_uint16 *)x->reg[**cp & 15];

			break;
		case OP_MOD+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] %
				*(_uint32 *)x->reg[**cp & 15];

			break;
		case OP_MOD+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] %
				*(_uint64 *)x->reg[**cp & 15];

			break;
		case OP_MODS:
			(*cp)++;
			*(char *)x->reg[**cp >> 4] = *(char *)x->reg[**cp >> 4] %
				*(char *)x->reg[**cp & 15];

			break;
		case OP_MODS+1:
			(*cp)++;
			*(_int16 *)x->reg[**cp >> 4] = *(_int16 *)x->reg[**cp >> 4] %
				*(_int16 *)x->reg[**cp & 15];

			break;
		case OP_MODS+2:
			(*cp)++;
			*(_int32 *)x->reg[**cp >> 4] = *(_int32 *)x->reg[**cp >> 4] %
				*(_int32 *)x->reg[**cp & 15];

			break;
		case OP_MODS+3:
			(*cp)++;
			*(_int64 *)x->reg[**cp >> 4] = *(_int64 *)x->reg[**cp >> 4] %
				*(_int64 *)x->reg[**cp & 15];

			break;
		case OP_SL:
			(*cp)++;
			*x->reg[**cp >> 4] <<= 1;

			break;
		case OP_SL+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] <<= 1;

			break;
		case OP_SL+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] <<= 1;

			break;
		case OP_SL+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] <<= 1;

			break;
		case OP_SLN:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] << *x->reg[**cp & 15];

			break;
		case OP_SLN+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] <<
				*x->reg[**cp & 15];

			break;
		case OP_SLN+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] <<
				*x->reg[**cp & 15];

			break;
		case OP_SLN+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] <<
				*x->reg[**cp & 15];

			break;
		case OP_SR:
			(*cp)++;
			*x->reg[**cp >> 4] >>= 1;

			break;
		case OP_SR+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] >>= 1;

			break;
		case OP_SR+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] >>= 1;

			break;
		case OP_SR+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] >>= 1;

			break;
		case OP_SRN:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] >> *x->reg[**cp & 15];

			break;
		case OP_SRN+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] >>
				*x->reg[**cp & 15];

			break;
		case OP_SRN+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] >>
				*x->reg[**cp & 15];

			break;
		case OP_SRN+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] >>
				*x->reg[**cp & 15];

			break;
		case OP_NEG:
			(*cp)++;
			*(char *)x->reg[**cp >> 4] = -*(char *)x->reg[**cp >> 4];

			break;
		case OP_NEG+1:
			(*cp)++;
			*(_int16 *)x->reg[**cp >> 4] = -*(_int16 *)x->reg[**cp >> 4];

			break;
		case OP_NEG+2:
			(*cp)++;
			*(_int32 *)x->reg[**cp >> 4] = -*(_int32 *)x->reg[**cp >> 4];

			break;
		case OP_NEG+3:
			(*cp)++;
			*(_int64 *)x->reg[**cp >> 4] = -*(_int64 *)x->reg[**cp >> 4];

			break;
		case OP_FNEG+2:
			(*cp)++;
			*(_float32 *)x->reg[**cp >> 4] = -*(_float32 *)x->reg[**cp >> 4];

			break;
		case OP_FNEG+3:
			(*cp)++;
			*(_float64 *)x->reg[**cp >> 4] = -*(_float64 *)x->reg[**cp >> 4];

			break;
		case OP_AND:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] & *x->reg[**cp & 15];

			break;
		case OP_AND+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] &
				*(_uint16 *)x->reg[**cp & 15];

			break;
		case OP_AND+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] &
				*(_uint32 *)x->reg[**cp & 15];

			break;
		case OP_AND+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] &
				*(_uint64 *)x->reg[**cp & 15];

			break;
		case OP_OR:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] | *x->reg[**cp & 15];

			break;
		case OP_OR+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] |
				*(_uint16 *)x->reg[**cp & 15];

			break;
		case OP_OR+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] |
				*(_uint32 *)x->reg[**cp & 15];

			break;
		case OP_OR+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] |
				*(_uint64 *)x->reg[**cp & 15];

			break;
		case OP_XOR:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] ^ *x->reg[**cp & 15];

			break;
		case OP_XOR+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] ^
				*(_uint16 *)x->reg[**cp & 15];

			break;
		case OP_XOR+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] ^
				*(_uint32 *)x->reg[**cp & 15];

			break;
		case OP_XOR+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] ^
				*(_uint64 *)x->reg[**cp & 15];

			break;
		case OP_NOT:
			(*cp)++;
			*x->reg[**cp >> 4] = ~*x->reg[**cp >> 4];

			break;
		case OP_NOT+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = ~*(_uint16 *)x->reg[**cp >> 4];

			break;
		case OP_NOT+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = ~*(_uint32 *)x->reg[**cp >> 4];

			break;
		case OP_NOT+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = ~*(_uint64 *)x->reg[**cp >> 4];

			break;
		case OP_BAND:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] && *x->reg[**cp & 15];

			break;
		case OP_BAND+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] &&
				*(_uint16 *)x->reg[**cp & 15];

			break;
		case OP_BAND+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] &&
				*(_uint32 *)x->reg[**cp & 15];

			break;
		case OP_BAND+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] &&
				*(_uint64 *)x->reg[**cp & 15];

			break;
		case OP_BOR:
			(*cp)++;
			*x->reg[**cp >> 4] = *x->reg[**cp >> 4] || *x->reg[**cp & 15];

			break;
		case OP_BOR+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = *(_uint16 *)x->reg[**cp >> 4] ||
				*(_uint16 *)x->reg[**cp & 15];

			break;
		case OP_BOR+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = *(_uint32 *)x->reg[**cp >> 4] ||
				*(_uint32 *)x->reg[**cp & 15];

			break;
		case OP_BOR+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = *(_uint64 *)x->reg[**cp >> 4] ||
				*(_uint64 *)x->reg[**cp & 15];

			break;
		case OP_BNOT:
			(*cp)++;
			*x->reg[**cp >> 4] = !*x->reg[**cp >> 4];

			break;
		case OP_BNOT+1:
			(*cp)++;
			*(_uint16 *)x->reg[**cp >> 4] = !*(_uint16 *)x->reg[**cp >> 4];

			break;
		case OP_BNOT+2:
			(*cp)++;
			*(_uint32 *)x->reg[**cp >> 4] = !*(_uint32 *)x->reg[**cp >> 4];

			break;
		case OP_BNOT+3:
			(*cp)++;
			*(_uint64 *)x->reg[**cp >> 4] = !*(_uint64 *)x->reg[**cp >> 4];

			break;
		case OP_CP:
			(*cp)++;
			i8_1 = *x->reg[**cp >> 4];
			i8_2 = *x->reg[**cp & 15];
			if (i8_1 == i8_2)
				x->z = 1;
			else
				x->z = 0;
			if (i8_1 < i8_2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_CP+1:
			(*cp)++;
			i16_1 = *(_int16 *)x->reg[**cp >> 4];
			i16_2 = *(_int16 *)x->reg[**cp & 15];
			if (i16_1 == i16_2)
				x->z = 1;
			else
				x->z = 0;
			if (i16_1 < i16_2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_CP+2:
			(*cp)++;
			i32_1 = *(_int32 *)x->reg[**cp >> 4];
			i32_2 = *(_int32 *)x->reg[**cp & 15];
			if (i32_1 == i32_2)
				x->z = 1;
			else
				x->z = 0;
			if (i32_1 < i32_2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_CP+3:
			(*cp)++;
			i64_1 = *(_int64 *)x->reg[**cp >> 4];
			i64_2 = *(_int64 *)x->reg[**cp & 15];
			if (i64_1 == i64_2)
				x->z = 1;
			else
				x->z = 0;
			if (i64_1 < i64_2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_SCP:
			(*cp)++;
			s8_1 = *x->reg[**cp >> 4];
			s8_2 = *x->reg[**cp & 15];
			if (s8_1 == s8_2)
				x->z = 1;
			else
				x->z = 0;
			if (s8_1 < s8_2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_SCP+1:
			(*cp)++;
			s16_1 = *(_int16 *)x->reg[**cp >> 4];
			s16_2 = *(_int16 *)x->reg[**cp & 15];
			if (s16_1 == s16_2)
				x->z = 1;
			else
				x->z = 0;
			if (s16_1 < s16_2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_SCP+2:
			(*cp)++;
			s32_1 = *(_int32 *)x->reg[**cp >> 4];
			s32_2 = *(_int32 *)x->reg[**cp & 15];
			if (s32_1 == s32_2)
				x->z = 1;
			else
				x->z = 0;
			if (s32_1 < s32_2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_SCP+3:
			(*cp)++;
			s64_1 = *(_int64 *)x->reg[**cp >> 4];
			s64_2 = *(_int64 *)x->reg[**cp & 15];
			if (s64_1 == s64_2)
				x->z = 1;
			else
				x->z = 0;
			if (s64_1 < s64_2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_FCP+2:
			(*cp)++;
			f32_1 = *(_float32 *)x->reg[**cp >> 4];
			f32_2 = *(_float32 *)x->reg[**cp & 15];
			if (f32_1 == f32_2)
				x->z = 1;
			else
				x->z = 0;
			if (f32_1 < f32_2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_FCP+3:
			(*cp)++;
			f64_1 = *(_float64 *)x->reg[**cp >> 4];
			f64_2 = *(_float64 *)x->reg[**cp & 15];
			if (f64_1 == f64_2)
				x->z = 1;
			else
				x->z = 0;
			if (f64_1 < f64_2)
				x->c = 1;
			else
				x->c = 0;

			break;
		case OP_TEST:
			(*cp)++;
			x->z = !*x->reg[**cp & 15];
			x->c = 0;

			break;
		case OP_TEST+1:
			(*cp)++;
			x->z = !*(_int16 *)x->reg[**cp & 15];
			x->c = 0;

			break;
		case OP_TEST+2:
			(*cp)++;
			x->z = !*(_int32 *)x->reg[**cp & 15];
			x->c = 0;

			break;
		case OP_TEST+3:
			(*cp)++;
			x->z = !*(_int64 *)x->reg[**cp & 15];
			x->c = 0;

			break;
		case OP_STEST:
			(*cp)++;
			x->z = !*(char *)x->reg[**cp & 15];
			x->c = *(char *)x->reg[**cp & 15] < 0;

			break;
		case OP_STEST+1:
			(*cp)++;
			x->z = !*(_int16 *)x->reg[**cp & 15];
			x->c = *(_int16 *)x->reg[**cp & 15] < 0;

			break;
		case OP_STEST+2:
			(*cp)++;
			x->z = !*(_int32 *)x->reg[**cp & 15];
			x->c = *(_int32 *)x->reg[**cp & 15] < 0;

			break;
		case OP_STEST+3:
			(*cp)++;
			x->z = !*(_int64 *)x->reg[**cp & 15];
			x->c = *(_int64 *)x->reg[**cp & 15] < 0;

			break;
		case OP_JP:
			*cp = code + *(_uint32 *)(*cp+1) - 1;

			break;
		case OP_JP_Z:
			if (x->z)
				*cp = code + *(_uint32 *)(*cp+1) - 1;
			else
				*cp += 4;

			break;
		case OP_JP_NZ:
			if (!x->z)
				*cp = code + *(_uint32 *)(*cp+1) - 1;
			else
				*cp += 4;

			break;
		case OP_JP_C:
			if (x->c)
				*cp = code + *(_uint32 *)(*cp+1) - 1;
			else
				*cp += 4;

			break;
		case OP_JP_NC:
			if (!x->c)
				*cp = code + *(_uint32 *)(*cp+1) - 1;
			else
				*cp += 4;

			break;
		case OP_JP_R:
			(*cp)++;
			*cp = code + *(_uint32 *)x->reg[**cp & 15] - 1;

			break;
		case OP_JP_R_Z:
			(*cp)++;
			if (x->z)
				*cp = code + *(_uint32 *)x->reg[**cp & 15] - 1;

			break;
		case OP_JP_R_NZ:
			(*cp)++;
			if (!x->z)
				*cp = code + *(_uint32 *)x->reg[**cp & 15] - 1;

			break;
		case OP_JP_R_C:
			(*cp)++;
			if (x->c)
				*cp = code + *(_uint32 *)x->reg[**cp & 15] - 1;

			break;
		case OP_JP_R_NC:
			(*cp)++;
			if (!x->c)
				*cp = code + *(_uint32 *)x->reg[**cp & 15] - 1;

			break;
		case OP_INC:
			(*cp)++;
			++*(char *)x->reg[**cp >> 4];

			break;
		case OP_INC+1:
			(*cp)++;
			++*(_int16 *)x->reg[**cp >> 4];

			break;
		case OP_INC+2:
			(*cp)++;
			++*(_int32 *)x->reg[**cp >> 4];

			break;
		case OP_INC+3:
			(*cp)++;
			++*(_int64 *)x->reg[**cp >> 4];

			break;
		case OP_FINC+2:
			(*cp)++;
			++*(_float32 *)x->reg[**cp >> 4];

			break;
		case OP_FINC+3:
			(*cp)++;
			++*(_float64 *)x->reg[**cp >> 4];

			break;
		case OP_DEC:
			(*cp)++;
			--*(char *)x->reg[**cp >> 4];

			break;
		case OP_DEC+1:
			(*cp)++;
			--*(_int16 *)x->reg[**cp >> 4];

			break;
		case OP_DEC+2:
			(*cp)++;
			--*(_int32 *)x->reg[**cp >> 4];

			break;
		case OP_DEC+3:
			(*cp)++;
			--*(_int64 *)x->reg[**cp >> 4];

			break;
		case OP_FDEC+2:
			(*cp)++;
			--*(_float32 *)x->reg[**cp >> 4];

			break;
		case OP_FDEC+3:
			(*cp)++;
			--*(_float64 *)x->reg[**cp >> 4];

			break;
		case OP_PUSH:
			(*cp)++;
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			s->n[0] = x->reg[**cp & 15][0];

			break;
		case OP_PUSH+1:
			(*cp)++;
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			s->n[0] = x->reg[**cp & 15][0];
			s->n[1] = x->reg[**cp & 15][1];

			break;
		case OP_PUSH+2:
			(*cp)++;
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			s->n[0] = x->reg[**cp & 15][0];
			s->n[1] = x->reg[**cp & 15][1];
			s->n[2] = x->reg[**cp & 15][2];
			s->n[3] = x->reg[**cp & 15][3];

			break;
		case OP_PUSH+3:
			(*cp)++;
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			memcpy(s->n, x->reg[**cp & 15], 8);

			break;
		case OP_POP:
			if (!x->stack)
			{
				xc_set_errno(XC_ERR_STACK_UNDERFLOW);
				return 0;
			}
			(*cp)++;
			x->reg[**cp & 15][0] = x->stack->n[0];
			s = x->stack;
			x->stack = x->stack->next;
			free(s);

			break;
		case OP_POP+1:
			if (!x->stack)
			{
				xc_set_errno(XC_ERR_STACK_UNDERFLOW);
				return 0;
			}
			(*cp)++;
			r1 = **cp & 15;
			x->reg[r1][0] = x->stack->n[0];
			x->reg[r1][1] = x->stack->n[1];
			s = x->stack;
			x->stack = x->stack->next;
			free(s);

			break;
		case OP_POP+2:
			if (!x->stack)
			{
				xc_set_errno(XC_ERR_STACK_UNDERFLOW);
				return 0;
			}
			(*cp)++;
			r1 = **cp & 15;
			x->reg[r1][0] = x->stack->n[0];
			x->reg[r1][1] = x->stack->n[1];
			x->reg[r1][2] = x->stack->n[2];
			x->reg[r1][3] = x->stack->n[3];
			s = x->stack;
			x->stack = x->stack->next;
			free(s);

			break;
		case OP_POP+3:
			if (!x->stack)
			{
				xc_set_errno(XC_ERR_STACK_UNDERFLOW);
				return 0;
			}
			(*cp)++;
			memcpy(x->reg[**cp & 15], x->stack->n, 8);
			s = x->stack;
			x->stack = x->stack->next;
			free(s);

			break;
		case OP_CALL_C:
			p = (pstack *)calloc(1, sizeof(pstack));
			p->next = x->cstack;
			x->cstack = p;
			x->cstack->ptr = *cp + 5;
			*cp = code + *(_uint32 *)(*cp+1) - 1;

			break;
		case OP_CALL_R:
			p = (pstack *)calloc(1, sizeof(pstack));
			p->next = x->cstack;
			x->cstack = p;
			x->cstack->ptr = *cp + 2;
			*cp = code + *(_uint32 *)x->reg[*(*cp+1) & 15] - 1;
			
			break;
		case OP_ARG_R:
			(*cp)++;

			if (++x->narg > x->marg)
				x->arg = (_int64 *)realloc(x->arg, sizeof(_int64) * (x->marg = x->narg));
			((char *)&x->arg[x->narg-1])[0] = x->reg[**cp & 15][0];
			memset(&((char *)&x->arg[x->narg-1])[1], 0, 7);

			break;
		case OP_ARG_R+1:
			(*cp)++;
			r1 = **cp & 15;

			if (++x->narg > x->marg)
				x->arg = (_int64 *)realloc(x->arg, sizeof(_int64) * (x->marg = x->narg));
			((char *)&x->arg[x->narg-1])[0] = x->reg[r1][0];
			((char *)&x->arg[x->narg-1])[1] = x->reg[r1][1];
			memset(&((char *)&x->arg[x->narg-1])[2], 0, 6);

			break;
		case OP_ARG_R+2:
			(*cp)++;
			
			if (++x->narg > x->marg)
				x->arg = (_int64 *)realloc(x->arg, sizeof(_int64) * (x->marg = x->narg));
			memcpy((char *)&x->arg[x->narg-1], x->reg[**cp & 15], 4);
			memset(&((char *)&x->arg[x->narg-1])[4], 0, 4);

			break;
		case OP_ARG_R+3:
			(*cp)++;
			
			if (++x->narg > x->marg)
				x->arg = (_int64 *)realloc(x->arg, sizeof(_int64) * (x->marg = x->narg));
			memcpy((char *)&x->arg[x->narg-1], x->reg[**cp & 15], 8);

			break;
		case OP_ARG_C:
			if (++x->narg > x->marg)
				x->arg = (_int64 *)realloc(x->arg, sizeof(_int64) * (x->marg = x->narg));
			((char *)&x->arg[x->narg-1])[0] = *++*cp;
			memset(&((char *)&x->arg[x->narg-1])[1], 0, 7);

			break;
		case OP_ARG_C+1:
			if (++x->narg > x->marg)
				x->arg = (_int64 *)realloc(x->arg, sizeof(_int64) * (x->marg = x->narg));
			((char *)&x->arg[x->narg-1])[0] = *++*cp;
			((char *)&x->arg[x->narg-1])[1] = *++*cp;
			memset(&((char *)&x->arg[x->narg-1])[2], 0, 6);

			break;
		case OP_ARG_C+2:
			if (++x->narg > x->marg)
				x->arg = (_int64 *)realloc(x->arg, sizeof(_int64) * (x->marg = x->narg));
			memcpy((char *)&x->arg[x->narg-1], ++*cp, 4);
			memset(&((char *)&x->arg[x->narg-1])[4], 0, 4);
			*cp += 3;

			break;
		case OP_ARG_C+3:
			if (++x->narg > x->marg)
				x->arg = (_int64 *)realloc(x->arg, sizeof(_int64) * (x->marg = x->narg));
			memcpy((char *)&x->arg[x->narg-1], ++*cp, 8);
			*cp += 7;

			break;
		case OP_HCALL_C:
			i16_1 = *(_uint16 *)++*cp;

			if (i16_1 >= n_hostfunc)
			{
				xc_set_errno(XC_ERR_HOSTSYM_SEGFAULT);
				return 0;
			}

			/*arg = calloc(narg = x->narg, sizeof(_int64));
			memcpy(arg, x->arg, sizeof(_int64) * narg);*/
			xc_set_errno(XC_ERR_NONE);
			i1 = hostfunc[i16_1](x, x->narg, x->arg);
//			call_hostsym((int)i16_1, x, narg, arg);
//			free(x->arg);
//			x->arg = malloc(0);
			x->narg = 0;
//			free(arg);
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;

			*(_uint64 *)s->n = i1;
			++*cp;

			break;
		case OP_HCALL_R:
			i16_1 = *(_uint16 *)x->reg[(*++*cp) & 15];

			if (i16_1 >= n_hostfunc)
			{
				xc_set_errno(XC_ERR_HOSTSYM_SEGFAULT);
				return 0;
			}

			/*arg = calloc(narg = x->narg, sizeof(_int64));
			memcpy(arg, x->arg, sizeof(_int64) * narg);*/
			i1 = hostfunc[i16_1](x, x->narg, x->arg);
//			i1 = call_hostsym((int)i16_1, x, narg, arg);
//			free(x->arg);
//			x->arg = malloc(0);
			x->narg = 0;
//			free(arg);
			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;

			*(_uint64 *)s->n = i1;

			break;
		case OP_RET:
			if (!x->cstack)
			{
				xc_set_errno(XC_ERR_NONE);
				return 0;
			}
			*cp = x->cstack->ptr - 1;
			p = x->cstack;
			x->cstack = p->next;
			free(p);

			break;
		case OP_RETV:
			if (!x->cstack)
			{
				xc_set_errno(XC_ERR_NONE);
				return 0;
			}

			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			s->n[0] = x->reg[(*++*cp) & 15][0];

			*cp = x->cstack->ptr - 1;
			p = x->cstack;
			x->cstack = p->next;
			free(p);

			break;
		case OP_RETV+1:
			if (!x->cstack)
			{
				xc_set_errno(XC_ERR_NONE);
				return 0;
			}

			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			r1 = (*++*cp) & 15;
			s->n[0] = x->reg[r1][0];
			s->n[1] = x->reg[r1][1];

			*cp = x->cstack->ptr - 1;
			p = x->cstack;
			x->cstack = p->next;
			free(p);

			break;
		case OP_RETV+2:
			if (!x->cstack)
			{
				xc_set_errno(XC_ERR_NONE);
				return 0;
			}

			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			memcpy(s->n, x->reg[(*++*cp) & 15], 4);

			*cp = x->cstack->ptr - 1;
			p = x->cstack;
			x->cstack = p->next;
			free(p);

			break;
		case OP_RETV+3:
			if (!x->cstack)
			{
				xc_set_errno(XC_ERR_NONE);
				return 0;
			}

			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;
			memcpy(s->n, x->reg[(*++*cp) & 15], 8);

			*cp = x->cstack->ptr - 1;
			p = x->cstack;
			x->cstack = p->next;
			free(p);

			break;
		case OP_CPY:
			(*cp)++;
			i32_1 = *(_int32 *)x->reg[**cp >> 4];
			i32_2 = *(_int32 *)x->reg[**cp & 15];
			(*cp)++;
			i32_3 = *(_int32 *)x->reg[**cp & 15];

			if (i32_1 + i32_3 > x->memsz || i32_2 + i32_3 > x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}

			memcpy(&(x->mem[i32_1]), &(x->mem[i32_2]), i32_3);

			break;
		case OP_SCPY:
			(*cp)++;
			i32_1 = *(_int32 *)x->reg[**cp >> 4];
			i32_2 = *(_int32 *)x->reg[**cp & 15];

			if (i32_1 >= x->memsz || i32_2 >= x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}

			strcpy(&(x->mem[i32_1]), &(x->mem[i32_2]));

			break;
		case OP_DCPY:
			(*cp)++;
			i32_1 = *(_int32 *)x->reg[**cp >> 4];
			i32_2 = *(_int32 *)x->reg[**cp & 15];

			if (i32_1 >= x->memsz || i32_2 >= (_uint32)x->parent->n_data)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}

			memcpy(&(x->mem[i32_1]), x->parent->data[i32_2].dat, x->parent->data[i32_2].sz);

			break;
		case OP_PMODE:
		case OP_IMODE:
		case OP_RMODE:
		case OP_OMODE:
			x->mode = **cp;
			s8 = (stack8 *)calloc(1, sizeof(stack8));
			s8->next = x->cmode;
			x->cmode = s8;
			s8->n = **cp;

			break;
		case OP_EMODE:
			if (x->cmode)
			{
				s8 = x->cmode;
				x->cmode = s8->next;
				free(s8);
				if (x->cmode)
					x->mode = x->cmode->n;
				else
					x->mode = OP_PMODE;
			}
			else
				x->mode = OP_PMODE;

			break;
		case OP_SYNC:
			if (x->serv)
			{
				++*cp;
				return -1;
			}

			break;
		case OP_RSEND:
			if (x->serv)
			{
				++*cp;
				if (x->serv->mode != XN_ALTER)
					x->serv->lmode = x->serv->mode;
				x->serv->mode = XN_TRANS;
				x->tmode = TRANS_RECV;
				x->transp = x->reg[**cp & 15];
				x->transz = 8;
				++*cp;

				do_recv(x);

				x->serv->mode = x->serv->lmode;
				x->tmode = TRANS_NONE;

				return -1;
			}
			else
				++*cp;

			break;
		case OP_MSEND:
			if (x->serv)
			{
				++*cp;
				if (x->serv->mode != XN_ALTER)
					x->serv->lmode = x->serv->mode;
				x->serv->mode = XN_TRANS;
				x->tmode = TRANS_RECV;
				i32_1 = *(_uint32 *)x->reg[**cp >> 4];
				i32_2 = *(_uint32 *)x->reg[**cp & 15];

				x->transp = &x->mem[i32_1];
				x->transz = (int)i32_2;
				++*cp;

				do_recv(x);

				x->serv->mode = x->serv->lmode;
				x->tmode = TRANS_NONE;

				return -1;
			}
			else
				++*cp;

			break;
		case OP_MBASE:
			s32_1 = *(_int32 *)++*cp;
			
			if ((_uint32)s32_1 > x->memsz)
			{
				xc_set_errno(XC_ERR_MODULE_MEMORY);
				return 0;
			}

			if (x->mb_base)
			{
				xc_set_errno(XC_ERR_MODULE_MEMORY);
				return 0;
			}

			mb = (memblock *)calloc(1, sizeof(memblock));
			mb->min = 0;
			mb->max = s32_1;

			x->mb_base = x->cur_mb = mb;
			*cp += 3;

			break;
		case OP_MALLOC:
			(*cp)++;
			s32_1 = *(_int32 *)x->reg[**cp & 15];

			if ((s32_2 = __xc_malloc(x, s32_1)) < 0)
				return 0;

			*(_uint32 *)x->reg[**cp >> 4] = s32_2;

			break;
		case OP_FREE:
			(*cp)++;

			__xc_free(x, *(_int32 *)x->reg[**cp & 15]);
			if (xc_get_errno())
				return 0;

			break;
		case OP_LALLOC:
			s32_1 = *(_int32 *)++*cp;
			*cp += 3;

			if ((s32_2 = __xc_malloc(x, s32_1)) < 0)
			{
				puts("Lalloc crashed!");
				return 0;
			}

			ln = (lint *)calloc(1, sizeof(lint));
			ln->next = x->local;
			if (x->local)
				x->local->prev = ln;
			x->local = ln;
			ln->n = s32_2;

			break;
		case OP_LFREE:
			if (!x->local)
			{
				xc_set_errno(XC_ERR_STACK_UNDERFLOW);
				return 0;
			}

			__xc_free(x, x->local->n);
			if (xc_get_errno())
			{
				puts("Lfree crashed!");
				return 0;
			}

			if (x->local->next)
				x->local->next->prev = NULL;
			ln = x->local;
			x->local = x->local->next;
			free(ln);

			break;
		case OP_LADDR:
			i32_1 = *(_int32 *)++*cp;
			*cp += 3;

			if (!x->local)
			{
				xc_set_errno(XC_ERR_STACK_UNDERFLOW);
				return 0;
			}

			s = (stackt *)calloc(1, sizeof(stackt));
			s->next = x->stack;
			x->stack = s;

			*(_int32 *)s->n = x->local->n + i32_1;

			break;
		case OP_CAST:
			{
				char t1, t2, sn1, sn2, sz1, sz2;
				_int64 mn1;
				_int32 ep1;
				unsigned char *sp1;

				if (!x->stack)
				{
					xc_set_errno(XC_ERR_STACK_UNDERFLOW);
					return 0;
				}

				i1 = *++*cp;
				t1 = (char)(i1 & 3);
				i1 >>= 2;
				sz1 = (char)(i1 & 3);
				sn1 = tsz[i1 & 3];
				i1 >>= 2;
				t2 = (char)(i1 & 3);
				i1 >>= 2;
				sz2 = (char)(i1 & 3);
				sn2 = tsz[i1 & 3];

//				printf("%d/%d vs. %d/%d\n", t1, sn1, t2, sn2);
				if (t1 == _C_FLOAT && t2 == _C_FLOAT)
				{
					read_float(x->stack->n, sz1, &mn1, &ep1);
					if (sn1 > sn2)
						mn1 >>= mantissa[sz1] - mantissa[sz2];
					else
						mn1 <<= mantissa[sz2] - mantissa[sz1];
					write_float(x->stack->n, mn1, ep1, sz2);
					read_float(x->stack->n, sz2, &mn1, &ep1);
				}
				else if (t1 == _C_FLOAT)
				{
					read_float(x->stack->n, sz1, &mn1, &ep1);
					write_int(x->stack->n, sz2, mn1 >> (mantissa[sz1] - ep1));
				}
				else if (t2 == _C_FLOAT)
				{
					read_int(x->stack->n, sz1, &mn1);
					write_float(x->stack->n, mn1, mantissa[sz2], sz2);
					read_float(x->stack->n, sz2, &mn1, &ep1);
				}
				else
				{
					if (t1 == _C_SIGNED)
					{
						if (sn2 > sn1)
						{
							if (x->stack->n[sn1-1] & 128)
							{
								i1 = 0;
								sp1 = x->stack->n;
								for (k = 0; k < sn1; ++k)
									i1 += (_int64)*sp1++ << (8 * k);
								i1 += imax[sz2] - imax[sz1];
								sp1 = x->stack->n;
								for (k = 0; k < sn2; ++k)
								{
									*sp1++ = (char)(i1 & 255);
									i1 >>= 8;
								}
							}
							else
								memset(&x->stack->n[sn1],
									0, sn2 - sn1);
						}
					}
					else if (sn2 > sn1)
						memset(&x->stack->n[sn1],
							0, sn2 - sn1);
				}

				*cp += 3;
			}

			break;
/*		case OP_CAST:
			{
				char t1, t2, sn1, sn2, sz1, sz2;

				if (!x->stack)
				{
					xc_set_errno(XC_ERR_STACK_UNDERFLOW);
					return 0;
				}

				i8_1 = *++*cp;
				t1 = (char)(i8_1 & 3);
				i8_1 >>= 2;
				sz1 = (char)(i8_1 & 3);
				sn1 = tsz[i8_1 & 3];
				i8_1 >>= 2;
				t2 = (char)(i8_1 & 3);
				i8_1 >>= 2;
				sz2 = (char)(i8_1 & 3);
				sn2 = tsz[i8_1 & 3];

//				printf("%d/%d vs. %d/%d\n", t1, sn1, t2, sn2);
				if (t1 == _C_FLOAT && t2 == _C_FLOAT)
				{
					if (sz1 == 2)
						*(_float64 *)x->stack->n =
							(_float64)*(_float32 *)x->stack->n;
					else
						*(_float32 *)x->stack->n =
							(_float32)*(_float64 *)x->stack->n;
				}
				else if (t1 == _C_FLOAT)
				{
					switch (sz2)
					{
					case 0:
						if (sz1 == 2)
							*(char *)x->stack->n =
								(char)*(_float32 *)x->stack->n;
						else
							*(char *)x->stack->n =
								(char)*(_float64 *)x->stack->n;
						break;
					case 1:
						if (sz1 == 2)
							*(_int16 *)x->stack->n =
								(_int16)*(_float32 *)x->stack->n;
						else
							*(_int16 *)x->stack->n =
								(_int16)*(_float64 *)x->stack->n;
						break;
					case 2:
						if (sz1 == 2)
							*(_int32 *)x->stack->n =
								(_int32)*(_float32 *)x->stack->n;
						else
							*(_int32 *)x->stack->n =
								(_int32)*(_float64 *)x->stack->n;
						break;
					case 3:
						if (sz1 == 2)
							*(_int64 *)x->stack->n =
								(_int64)*(_float32 *)x->stack->n;
						else
							*(_int64 *)x->stack->n =
								(_int64)*(_float64 *)x->stack->n;
						break;
					}
				}
				else if (t2 == _C_FLOAT)
				{
					if (sz2 == 2)
					{
						switch (sz1)
						{
						case 0:
							*(_float32 *)x->stack->n =
								(_float32)*(char *)x->stack->n;
							break;
						case 1:
							*(_float32 *)x->stack->n =
								(_float32)*(_int16 *)x->stack->n;
							break;
						case 2:
							*(_float32 *)x->stack->n =
								(_float32)*(_int32 *)x->stack->n;
							break;
						case 3:
							*(_float32 *)x->stack->n =
								(_float32)*(_int64 *)x->stack->n;
							break;
						}
					}
					else
					{
						switch (sz1)
						{
						case 0:
							*(_float64 *)x->stack->n =
								(_float64)*(char *)x->stack->n;
							break;
						case 1:
							*(_float64 *)x->stack->n =
								(_float64)*(_int16 *)x->stack->n;
							break;
						case 2:
							*(_float64 *)x->stack->n =
								(_float64)*(_int32 *)x->stack->n;
							break;
						case 3:
							*(_float64 *)x->stack->n =
								(_float64)*(_int64 *)x->stack->n;
							break;
						}
					}
				}
				else
				{
					if (t1 == _C_SIGNED)
					{
						if (t2 == _C_SIGNED)
						{
							switch (sz2)
							{
							case 0:
								switch (sz1)
								{
								case 1:
									*(char *)x->stack->n =
										(char)*(_int16 *)x->stack->n;
									break;
								case 2:
									*(char *)x->stack->n =
										(char)*(_int32 *)x->stack->n;
									break;
								case 3:
									*(char *)x->stack->n =
										(char)*(_int64 *)x->stack->n;
									break;
								}
								break;
							case 1:
								switch (sz1)
								{
								case 0:
									*(_int16 *)x->stack->n =
										(_int16)*(char *)x->stack->n;
									break;
								case 2:
									*(_int16 *)x->stack->n =
										(_int16)*(_int32 *)x->stack->n;
									break;
								case 3:
									*(_int16 *)x->stack->n =
										(_int16)*(_int64 *)x->stack->n;
									break;
								}
								break;
							case 2:
								switch (sz1)
								{
								case 0:
									*(_int32 *)x->stack->n =
										(_int32)*(char *)x->stack->n;
									break;
								case 1:
									*(_int32 *)x->stack->n =
										(_int32)*(_int16 *)x->stack->n;
									break;
								case 3:
									*(_int32 *)x->stack->n =
										(_int32)*(_int64 *)x->stack->n;
									break;
								}
								break;
							case 3:
								switch (sz1)
								{
								case 0:
									*(_int64 *)x->stack->n =
										(_int64)*(char *)x->stack->n;
									break;
								case 1:
									*(_int64 *)x->stack->n =
										(_int64)*(_int16 *)x->stack->n;
									break;
								case 2:
									*(_int64 *)x->stack->n =
										(_int64)*(_int32 *)x->stack->n;
									break;
								}
								break;
							}
						}
						else
						{
							switch (sz2)
							{
							case 0:
								switch (sz1)
								{
								case 1:
									*(char *)x->stack->n =
										(char)*(_uint16 *)x->stack->n;
									break;
								case 2:
									*(char *)x->stack->n =
										(char)*(_uint32 *)x->stack->n;
									break;
								case 3:
									*(char *)x->stack->n =
										(char)*(_uint64 *)x->stack->n;
									break;
								}
								break;
							case 1:
								switch (sz1)
								{
								case 0:
									*(_int16 *)x->stack->n =
										(_int16)*x->stack->n;
									break;
								case 2:
									*(_int16 *)x->stack->n =
										(_int16)*(_uint32 *)x->stack->n;
									break;
								case 3:
									*(_int16 *)x->stack->n =
										(_int16)*(_uint64 *)x->stack->n;
									break;
								}
								break;
							case 2:
								switch (sz1)
								{
								case 0:
									*(_int32 *)x->stack->n =
										(_int32)*x->stack->n;
									break;
								case 1:
									*(_int32 *)x->stack->n =
										(_int32)*(_uint16 *)x->stack->n;
									break;
								case 3:
									*(_int32 *)x->stack->n =
										(_int32)*(_uint64 *)x->stack->n;
									break;
								}
								break;
							case 3:
								switch (sz1)
								{
								case 0:
									*(_int64 *)x->stack->n =
										(_int64)*x->stack->n;
									break;
								case 1:
									*(_int64 *)x->stack->n =
										(_int64)*(_uint16 *)x->stack->n;
									break;
								case 2:
									*(_int64 *)x->stack->n =
										(_int64)*(_uint32 *)x->stack->n;
									break;
								}
								break;
							}
						}
					}
					else
					{
						if (t2 == _C_SIGNED)
						{
							switch (sz2)
							{
							case 0:
								switch (sz1)
								{
								case 1:
									*x->stack->n =
										(unsigned char)*(_int16 *)x->stack->n;
									break;
								case 2:
									*x->stack->n =
										(unsigned char)*(_int32 *)x->stack->n;
									break;
								case 3:
									*x->stack->n =
										(unsigned char)*(_int64 *)x->stack->n;
									break;
								}
								break;
							case 1:
								switch (sz1)
								{
								case 0:
									*(_uint16 *)x->stack->n =
										(_uint16)*(char *)x->stack->n;
									break;
								case 2:
									*(_uint16 *)x->stack->n =
										(_uint16)*(_int32 *)x->stack->n;
									break;
								case 3:
									*(_uint16 *)x->stack->n =
										(_uint16)*(_int64 *)x->stack->n;
									break;
								}
								break;
							case 2:
								switch (sz1)
								{
								case 0:
									*(_uint32 *)x->stack->n =
										(_uint32)*(char *)x->stack->n;
									break;
								case 1:
									*(_uint32 *)x->stack->n =
										(_uint32)*(_int16 *)x->stack->n;
									break;
								case 3:
									*(_uint32 *)x->stack->n =
										(_uint32)*(_int64 *)x->stack->n;
									break;
								}
								break;
							case 3:
								switch (sz1)
								{
								case 0:
									*(_uint64 *)x->stack->n =
										(_uint64)*(char *)x->stack->n;
									break;
								case 1:
									*(_uint64 *)x->stack->n =
										(_uint64)*(_int16 *)x->stack->n;
									break;
								case 2:
									*(_uint64 *)x->stack->n =
										(_uint64)*(_int32 *)x->stack->n;
									break;
								}
								break;
							}
						}
						else
						{
							switch (sz2)
							{
							case 0:
								switch (sz1)
								{
								case 1:
									*x->stack->n =
										(unsigned char)*(_uint16 *)x->stack->n;
									break;
								case 2:
									*x->stack->n =
										(unsigned char)*(_uint32 *)x->stack->n;
									break;
								case 3:
									*x->stack->n =
										(unsigned char)*(_uint64 *)x->stack->n;
									break;
								}
								break;
							case 1:
								switch (sz1)
								{
								case 0:
									*(_uint16 *)x->stack->n =
										(_uint16)*x->stack->n;
									break;
								case 2:
									*(_uint16 *)x->stack->n =
										(_uint16)*(_uint32 *)x->stack->n;
									break;
								case 3:
									*(_uint16 *)x->stack->n =
										(_uint16)*(_uint64 *)x->stack->n;
									break;
								}
								break;
							case 2:
								switch (sz1)
								{
								case 0:
									*(_uint32 *)x->stack->n =
										(_uint32)*x->stack->n;
									break;
								case 1:
									*(_uint32 *)x->stack->n =
										(_uint32)*(_uint16 *)x->stack->n;
									break;
								case 3:
									*(_uint32 *)x->stack->n =
										(_uint32)*(_uint64 *)x->stack->n;
									break;
								}
								break;
							case 3:
								switch (sz1)
								{
								case 0:
									*(_uint64 *)x->stack->n =
										(_uint64)*x->stack->n;
									break;
								case 1:
									*(_uint64 *)x->stack->n =
										(_uint64)*(_uint16 *)x->stack->n;
									break;
								case 2:
									*(_uint64 *)x->stack->n =
										(_uint64)*(_uint32 *)x->stack->n;
									break;
								}
								break;
							}
						}
					}
				}

				*cp += 3;
			}

			break;*/
		case OP_GBF:
			r1 = (*++*cp) & 15;
			i32_1 = *(_int32 *)x->reg[r1];
			i8_1 = *++*cp;
			i8_2 = *++*cp;

			i32_1 >>= i8_1;
			*(_int32 *)x->reg[r1] = i32_1 & ((1 << i8_2) - 1);

			break;
		case OP_SBF:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			i32_1 = *(_int32 *)x->reg[r1];
			i32_2 = *(_int32 *)x->reg[r2];
			i8_1 = *++*cp;
			i8_2 = *++*cp;

			i32_2 &= (1 << i8_2) - 1;
			i32_2 <<= i8_1;
			for (ku = 0; ku < i8_2; ++ku)
				if (i32_1 & (1 << (i8_1 + ku)))
					i32_1 ^= 1 << (i8_1 + ku);
			i32_1 |= i32_2;

			*(_int32 *)x->reg[r1] = i32_1;

			break;
		case OP_SGBF:
			r1 = (*++*cp) & 15;
			i32_1 = *(_int32 *)x->reg[r1];
			i8_1 = *++*cp;
			i8_2 = *++*cp;

			i32_1 >>= i8_1;
			i32_1 &= ((1 << i8_2) - 1);

			if (i32_1 & (1 << (i8_2-1)))
				i32_1 = 0xFFFF - ((0xFFFF & ((1 << i8_2) - 1)) - i32_1);

			*(_int32 *)x->reg[r1] = i32_1;

			break;
		case OP_SSBF:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			i32_1 = *(_int32 *)x->reg[r1];
			i32_2 = *(_int32 *)x->reg[r2];
			i8_1 = *++*cp;
			i8_2 = *++*cp;

			if (i32_2 & 0x8000)
				i32_2 = (0xFFFF & ((1 << i8_2) - 1)) - (0x8000 - i32_2);

			i32_2 &= (1 << i8_2) - 1;
			i32_2 <<= i8_1;
			for (ku = 0; ku < i8_2; ++ku)
				if (i32_1 & (1 << (i8_1 + ku)))
					i32_1 ^= 1 << (i8_1 + ku);
			i32_1 |= i32_2;

			*(_int32 *)x->reg[r1] = i32_1;

			break;
		case OP_ZERO:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			i32_1 = *(_int32 *)x->reg[r1];
			i32_2 = *(_int32 *)x->reg[r2];

			if ((i32_1 + i32_2) < 0 || (i32_1 + i32_2) >= x->memsz)
			{
				xc_set_errno(XC_ERR_SEGFAULT);
				return 0;
			}

			memset(&x->mem[i32_1], 0, i32_2);

			break;
		case OP_SLEN:
			(*cp)++;
			r1 = **cp >> 4;
			r2 = **cp & 15;

			i32_1 = *(_int32 *)x->reg[r2];

			if (!i32_1)
			{
				memset(x->reg[r1], 0, 4);
				break;
			}

			i32_2 = 1;
			for (; ; ++i32_1)
			{
				if (i32_1 >= x->memsz)
				{
					xc_set_errno(XC_ERR_SEGFAULT);
					return 0;
				}
				if (!x->mem[i32_1])
					break;
				++i32_2;
			}

			*(_int32 *)x->reg[r1] = i32_2;

			break;
		case OP_PING:
			if (x->serv)
				xnet_sendc(x->serv, (char)OP_PONG);

			break;
#endif
		}
	++*cp;

	return 1;
}

static do_xinst(xinst *x)
{
	int n, ycount = 0;

	xc_set_errno(0);
	c_inst_p = x;

	while (!_abort_flag)
	{
		if (++ycount >= YCOUNT)
		{
			YIELD;
			ycount = 0;
		}
		if (!x->cp)
			return 1;
		if (x->cp > x->code_end)
		{
			xc_set_errno(0);
			return 0;
		}
		if (!(n = do_op(x, &x->cp, x->code)))
			return 0;
		else if (n < 0)
			return 1;
	}

	return 1;
}

static do_xinst_net(xinst *x)
{
	int j, n, ycount = 0;
	unsigned char nbuf[S_NBUF], op;
	unsigned char **cp, *code, *code_end, *opp;
	char sync = 0, dead = 0;
	stack8 *s8;

	xc_set_errno(0);
	c_inst_p = x;

	while (!_abort_flag)
	{
		if (++ycount >= YCOUNT)
		{
			YIELD;
			ycount = 0;
		}

		if (x->serv && sync)
		{
			if (!x->nbufp || dead)
				n = xnet_recv(x->serv, nbuf, S_NBUF - x->nbufp);
			else
				n = 0;

			if (!n && xc_get_errno())
			{
				xc_set_errno(0);
				return 0;
			}
			else if (n)
			{
				xc_set_errno(0);
				if (n != -1 || x->nbufp)
				{
					if (n == -1) n = 0;
					active_inst = 1;

					memcpy(&x->nbuf[x->nbufp], nbuf, n);
					n += x->nbufp;

					dead = 1;
					for (j = 0; j < n; ++j)
					{
						if (x->nbuf[j] >= N_OPS)
							continue;
						if (n - j >= opsz[x->nbuf[j]])
						{
							dead = 0;
							op = x->nbuf[j];
							if (x->s_code_end - x->s_code + opsz[op] >=
								x->sz)
							{
								// Cycle buffer
								int sz = x->s_code_end - x->s_cp + 1;
								x->s_code_end -= (x->s_cp - x->s_code);
								memmove(x->s_code, x->s_cp, sz);
								x->s_cp = x->s_code;
							}
							memcpy(x->s_code_end+1, &x->nbuf[j], opsz[op]);
							x->s_code_end += opsz[op];
							j += opsz[op] - 1;
							if (!x->s_cp)
								x->s_cp = x->s_code;
							if (op == OP_RSEND || op == OP_MSEND)
							{
								++j;
								break;
							}
						}
						else
							break;
					}
					if (j < n)
					{
						memcpy(nbuf, &x->nbuf[j], n - j);
						memcpy(x->nbuf, nbuf, n - j);
					}
					x->nbufp = n - j;
				}
			}
		}

		if (sync)
		{
			cp = &x->s_cp;
			code = x->s_code;
			code_end = x->s_code_end;
		}
		else
		{
			cp = &x->cp;
			code = x->code;
			code_end = x->code_end;
		}

		if (!*cp)
			return 1;
		if (*cp > code_end)
		{
			if (sync)
				continue;
			else
			{
				xc_set_errno(0);
				return 0;
			}
		}

		if (x->mode != OP_PMODE && !(x->mode == OP_IMODE && !x->in) && !(x->mode == OP_OMODE && !x->out) && !(x->mode == OP_RMODE && !x->serv) && !sync)
		{
			op = **cp;
			opp = *cp + 1;

			if (x->cmode)
				if (x->cmode->n == x->mode && op == OP_EMODE)
					goto Regular;

			switch (x->mode)
			{
				case OP_IMODE:
					xnet_send(x->in, *cp, opsz[op]);
					*cp += opsz[op];
					xc_set_errno(0);
					break;
				case OP_RMODE:
					xnet_send(x->serv, *cp, opsz[op]);
					*cp += opsz[op];
					xc_set_errno(0);
					break;
				case OP_OMODE:
					xnet_send(x->out, *cp, opsz[op]);
					*cp += opsz[op];
					xc_set_errno(0);
					break;
			}

			if (op == OP_SYNC && x->serv)
			{
				sync = 1;
				continue;
			}
			else if (op >= OP_PMODE && op <= OP_OMODE)
			{
				s8 = (stack8 *)calloc(1, sizeof(stack8));
				s8->next = x->cmode;
				x->cmode = s8;
				s8->n = op;
			}
			else if (op == OP_EMODE)
			{
				if (x->cmode)
				{
					s8 = x->cmode;
					x->cmode = s8->next;
					free(s8);
					if (!x->cmode)
						x->mode = OP_PMODE;
				}
			}
			else if (op == OP_RSEND)
			{
				if (x->mode == OP_RMODE)
				{
					x->transp = x->reg[*opp & 15];
					x->transz = 8;

					xnet_send(x->serv, x->transp, x->transz);

					continue;
				}
				else
				{
					sync = 1;
					continue;
				}
			}
			else if (op == OP_MSEND)
			{
				if (x->mode == OP_RMODE)
				{
					x->transp = &x->mem[*(_int32 *)x->reg[*opp >> 4]];
					x->transz = *(_int32 *)x->reg[*opp & 15];

					xnet_send(x->serv, x->transp, x->transz);

					continue;
				}
				else
				{
					sync = 1;
					continue;
				}
			}
			continue;
		}
Regular:
		if (!(n = do_op(x, cp, code)))
			return 0;
		else if (n < 0)
			sync = 0;
	}

	return 1;
}

static do_xinst_streamed(xinst *x)
{
	int j, n, ycount = 0;
	unsigned char nbuf[S_NBUF], op;
	unsigned char *opp;
	char sync = 0, dead = 0;
	stack8 *s8;

	xc_set_errno(0);
	c_inst_p = x;

	while (!_abort_flag)
	{
		if (++ycount >= YCOUNT)
		{
			YIELD;
			ycount = 0;
		}

		if (!x->nbufp || dead)
		{
			n = xnet_recv(x->serv, nbuf, S_NBUF - x->nbufp);

			if (!n && xc_get_errno())
			{
				xc_set_errno(0);
				return 0;
			}
		}
		else
			n = 0;

		xc_set_errno(0);
		if (n == -1) n = 0;
		active_inst = 1;

		memcpy(&x->nbuf[x->nbufp], nbuf, n);
		n += x->nbufp;

		dead = 1;
		for (j = 0; j < n; ++j)
		{
			if (x->nbuf[j] >= N_OPS)
				continue;
			if (n - j >= opsz[x->nbuf[j]])
			{
				dead = 0;
				op = x->nbuf[j];
				if (x->s_code_end - x->s_code + opsz[op] >=
					x->sz)
				{
					// Cycle buffer
					int sz = x->s_code_end - x->s_cp + 1;
					x->s_code_end -= (x->s_cp - x->s_code);
					memmove(x->s_code, x->s_cp, sz);
					x->s_cp = x->s_code;
				}
				memcpy(x->s_code_end+1, &x->nbuf[j], opsz[op]);
				x->s_code_end += opsz[op];
				j += opsz[op] - 1;
				if (!x->s_cp)
					x->s_cp = x->s_code;
				if (op == OP_RSEND || op == OP_MSEND)
				{
					++j;
					break;
				}
			}
			else
				break;
		}
		if (j < n)
			memmove(x->nbuf, &x->nbuf[j], n - j);
		x->nbufp = n - j;

		while (!_abort_flag && x->s_cp && x->s_cp <= x->s_code_end)
		{
			if (x->mode != OP_PMODE && !(x->mode == OP_IMODE && !x->in) && !(x->mode == OP_OMODE && !x->out) && !(x->mode == OP_RMODE && !x->serv) && !sync)
			{
				op = *x->s_cp;
				opp = x->s_cp + 1;

				if (x->cmode)
					if (x->cmode->n == x->mode && op == OP_EMODE)
						goto Regular2;

				switch (x->mode)
				{
					case OP_IMODE:
						xnet_send(x->in, x->s_cp, opsz[*x->s_cp]);
						x->s_cp += opsz[op];
						xc_set_errno(0);
						break;
					case OP_RMODE:
						xnet_send(x->serv, x->s_cp, opsz[*x->s_cp]);
						x->s_cp += opsz[op];
						xc_set_errno(0);
						break;
					case OP_OMODE:
						xnet_send(x->out, x->s_cp, opsz[*x->s_cp]);
						x->s_cp += opsz[op];
						xc_set_errno(0);
						break;
				}

				if (op == OP_SYNC && x->serv)
				{
					sync = 1;
					continue;
				}
				else if (op >= OP_PMODE && op <= OP_OMODE)
				{
	//				printf("Streaming mode %u\n", (unsigned)op);
					s8 = (stack8 *)calloc(1, sizeof(stack8));
					s8->next = x->cmode;
					x->cmode = s8;
					s8->n = op;
				}
				else if (op == OP_EMODE)
				{
	//				puts("Streaming mode end");
					if (x->cmode)
					{
						s8 = x->cmode;
						x->cmode = s8->next;
						free(s8);
						if (!x->cmode)
							x->mode = OP_PMODE;
					}
				}
				else if (op == OP_RSEND)
				{
					if (x->mode == OP_RMODE)
					{
						x->transp = x->reg[*opp & 15];
						x->transz = 8;

						xnet_send(x->serv, x->transp, x->transz);

						continue;
					}
					else
					{
						sync = 1;
						continue;
					}
				}
				else if (op == OP_MSEND)
				{
					if (x->mode == OP_RMODE)
					{
						x->transp = &x->mem[*(_int32 *)x->reg[*opp >> 4]];
						x->transz = *(_int32 *)x->reg[*opp & 15];

						xnet_send(x->serv, x->transp, x->transz);

						continue;
					}
					else
					{
						sync = 1;
						continue;
					}
				}
				continue;
			}
Regular2:
			op = *x->s_cp;
			if (!(n = do_op(x, &x->s_cp, x->s_code)))
				return 0;
			else if (n < 0)
				sync = 0;
		}
	}

	return 1;
}