// xNet network API for VM

#include "xc.h"
#include "xctypes.h"
#include "xcvars.h"

int sil;
unsigned long block;

static is_ip(const char *str)
{
	for (; *str; ++str)
		if (*str != '.' && *str != ' ' && !isdigit(*str))
			return 0;
	return 1;
}

int xnet_init()
{
	sil = sizeof(SOCKADDR_IN);

	return 1;
}

int xnet_create(xnet *x, int type)
{
	WSADATA wsaData;
	static wsa_init = 0;

	if (type < 1 || type > XN_TYPEMAX)
	{
		xc_set_errno(XN_ERR_BAD_TYPE);
		return 0;
	}

	switch (x->type = type)
	{
		case XN_TCP:
		case XN_TCP_BLOCKING:
		case XN_UDP:
			if (!wsa_init)
			{
				if (WSAStartup(MAKEWORD(1, 1), &wsaData))
				{
					xc_set_errno(XN_ERR_LIBRARIES);
					return 0;
				}
				if (LOBYTE(wsaData.wVersion)!= 1 || HIBYTE(wsaData.wVersion)!= 1)
				{
					xc_set_errno(XN_ERR_LIBRARIES);
					WSACleanup();
					return 0;
				}
				wsa_init = 1;
			}
			break;
	}

	switch (type)
	{
		case XN_TCP:
			if ((x->info.tcp.sock = socket(AF_INET, SOCK_STREAM, 0)) == SOCKET_ERROR)
			{
				xc_set_errno(XN_ERR_CREATE);
				return 0;
			}
			block = 1;
			ioctlsocket(x->info.tcp.sock, FIONBIO, &block);
			break;
		case XN_TCP_BLOCKING:
			if ((x->info.tcp.sock = socket(AF_INET, SOCK_STREAM, 0)) == SOCKET_ERROR)
			{
				xc_set_errno(XN_ERR_CREATE);
				return 0;
			}
			break;
		case XN_UDP:
			if ((x->info.tcp.sock = socket(AF_INET, SOCK_DGRAM, 0)) == SOCKET_ERROR)
			{
				xc_set_errno(XN_ERR_CREATE);
				return 0;
			}
			block = 1;
			ioctlsocket(x->info.tcp.sock, FIONBIO, &block);
			break;
	}

	x->type = type;
	return 1;
}

int xnet_connect(xnet *x, ...)
{
	va_list ap;
	char *addr;
	int port;

	switch (x->type)
	{
		case XN_TCP:
		case XN_TCP_BLOCKING:
			va_start(ap, x);
			addr = va_arg(ap, char*);
			port = va_arg(ap, int);
			va_end(ap);

			memset((char *)&x->info.tcp.remote, 0, sizeof(SOCKADDR_IN));
			x->info.tcp.remote.sin_family = AF_INET;
			if (is_ip(addr))
				x->info.tcp.remote.sin_addr.s_addr = inet_addr(addr);
			else
			{
				struct hostent *h;

				if (!(h = gethostbyname(addr)))
				{
					xc_set_errno(XN_ERR_RESOLVE);
					return 0;
				}

				x->info.tcp.remote.sin_addr.s_addr =
					*((u_long *)h->h_addr);
			}

			x->info.tcp.remote.sin_port = htons((u_short)port);

			if (x->type == XN_TCP)
			{
				block = 0;
				ioctlsocket(x->info.tcp.sock, FIONBIO, &block);
			}
			if (connect(x->info.tcp.sock, (SOCKADDR*)&x->info.tcp.remote, sizeof(SOCKADDR_IN)) == SOCKET_ERROR)
			{
				xc_set_errno(XN_ERR_CONNECT);
				return 0;
			}
			if (x->type == XN_TCP)
			{
				block = 1;
				ioctlsocket(x->info.tcp.sock, FIONBIO, &block);
			}

			break;
		case XN_UDP:
			va_start(ap, x);
			addr = va_arg(ap, char*);
			port = va_arg(ap, int);
			va_end(ap);

			memset((char *)&x->info.tcp.remote, 0, sizeof(SOCKADDR_IN));
			x->info.tcp.remote.sin_family = AF_INET;
			if (is_ip(addr))
				x->info.tcp.remote.sin_addr.s_addr = inet_addr(addr);
			else
			{
				struct hostent *h;

				if (!(h = gethostbyname(addr)))
				{
					xc_set_errno(XN_ERR_RESOLVE);
					return 0;
				}

				x->info.tcp.remote.sin_addr.s_addr =
					*((u_long *)h->h_addr);
			}

			x->info.tcp.remote.sin_port = htons((u_short)port);
			break;
		default:
			xc_set_errno(XN_ERR_UNSUPPORTED);
			return 0;
	}

	x->mode = XN_CON;
	return 1;
}

int xnet_listen(xnet *x, ...)
{
	va_list ap;
	int port, maxcon;

	switch (x->type)
	{
		case XN_TCP:
		case XN_TCP_BLOCKING:
			va_start(ap, x);
			port = va_arg(ap, int);
			maxcon = va_arg(ap, int);
			va_end(ap);

			memset((char *)&(x->info.tcp.local), 0, sizeof(SOCKADDR_IN)); 
			x->info.tcp.local.sin_family = AF_INET;
			x->info.tcp.local.sin_addr.s_addr = htonl(INADDR_ANY);
			x->info.tcp.local.sin_port = htons((u_short)port);

			if (bind(x->info.tcp.sock, (SOCKADDR*)&x->info.tcp.local, sizeof(SOCKADDR)) == SOCKET_ERROR)
			{
				xc_set_errno(XN_ERR_BIND);
				return 0;
			}

			if (listen(x->info.tcp.sock, maxcon) == SOCKET_ERROR)
			{
				xc_set_errno(XN_ERR_LISTEN);
				return 0;
			}

			break;
		case XN_UDP:
			va_start(ap, x);
			port = va_arg(ap, int);
			va_end(ap);

			memset((char *)&(x->info.tcp.local), 0, sizeof(SOCKADDR_IN)); 
			x->info.tcp.local.sin_family = AF_INET;
			x->info.tcp.local.sin_addr.s_addr = htonl(INADDR_ANY);
			x->info.tcp.local.sin_port = htons((u_short)port);

			if (bind(x->info.tcp.sock, (SOCKADDR*)&x->info.tcp.local, sizeof(SOCKADDR)) == SOCKET_ERROR)
			{
				xc_set_errno(XN_ERR_BIND);
				return 0;
			}

			break;
		default:
			xc_set_errno(XN_ERR_UNSUPPORTED);
			return 0;
	}

	x->mode = XN_LISTEN;
	return 1;
}

int xnet_accept(xnet *l, xnet *a)
{
	int n;

	switch (l->type)
	{
		case XN_TCP:
		case XN_TCP_BLOCKING:
			if (a->type != XN_TCP && a->type != XN_TCP_BLOCKING)
			{
				xc_set_errno(XN_ERR_TYPE_COMPAT);
				return 0;
			}

			memset(&a->info.tcp.remote, 0, sizeof(SOCKADDR_IN));
			if ((n = a->info.tcp.sock = accept(l->info.tcp.sock, (SOCKADDR*)&a->info.tcp.remote, &sil)) == INVALID_SOCKET)
			{
				if (WSAGetLastError() == WSAEWOULDBLOCK)
					return -1;
				xc_set_errno(XN_ERR_ACCEPT);
				return 0;
			}
			else if (n == SOCKET_ERROR)
			{
				xc_set_errno(XN_ERR_ACCEPT);
				return 0;
			}
			break;
		default:
			xc_set_errno(XN_ERR_UNSUPPORTED);
			return 0;
	}

	a->mode = XN_CON;
	return 1;
}

int xnet_close(xnet *x)
{
	switch (x->type)
	{
		case XN_TCP:
		case XN_TCP_BLOCKING:
		case XN_UDP:
			closesocket(x->info.tcp.sock);
			break;
		default:
			xc_set_errno(XN_ERR_UNSUPPORTED);
			return 0;
	}

	x->mode = XN_DISCON;
	return 1;
}

int xnet_send(xnet *x, char *str, int len)
{
	switch (x->type)
	{
		case XN_TCP:
		case XN_TCP_BLOCKING:
			if (x->type == XN_TCP)
			{
				block = 0;
				ioctlsocket(x->info.tcp.sock, FIONBIO, &block);
			}
			if (send(x->info.tcp.sock, str, len, 0) == SOCKET_ERROR)
			{
				xc_set_errno(XN_ERR_SEND);
				return 0;
			}
			if (x->type == XN_TCP)
			{
				block = 1;
				ioctlsocket(x->info.tcp.sock, FIONBIO, &block);
			}
			break;
		case XN_UDP:
			block = 0;
			ioctlsocket(x->info.tcp.sock, FIONBIO, &block);
			if (sendto(x->info.tcp.sock, str, len, 0,
				(SOCKADDR*)&x->info.tcp.remote, sizeof(SOCKADDR))
				== SOCKET_ERROR)
			{
				xc_set_errno(XN_ERR_SEND);
				return 0;
			}
			block = 1;
			ioctlsocket(x->info.tcp.sock, FIONBIO, &block);
			break;
		default:
			xc_set_errno(XN_ERR_UNSUPPORTED);
			return 0;
	}

	xc_set_errno(0);
	return 1;
}

int xnet_sends(xnet *x, char *str)
{
	return xnet_send(x, str, strlen(str));
}

int xnet_sendc(xnet *x, char ch)
{
	return xnet_send(x, &ch, 1);
}

int xnet_recv(xnet *x, char *buf, int len)
{
	int n;

	switch (x->type)
	{
		case XN_TCP:
		case XN_TCP_BLOCKING:
			if ((n = recv(x->info.tcp.sock, buf, len, 0)) == SOCKET_ERROR)
			{
				if (WSAGetLastError() == WSAEWOULDBLOCK)
				{
					xc_set_errno(XN_ERR_BLOCKING);
					return -1;
				}
				xc_set_errno(XN_ERR_RECV);
				return 0;
			}
			else if (!n)
			{
				x->mode = XN_DISCON;
				xc_set_errno(XN_ERR_CLOSED);
				return 0;
			}
			break;
		case XN_UDP:
			if ((n = recvfrom(x->info.tcp.sock, buf, len, 0,
				(SOCKADDR*)&x->info.tcp.remote, &sil))
				== SOCKET_ERROR)
			{
				if (WSAGetLastError() == WSAEWOULDBLOCK)
				{
					xc_set_errno(XN_ERR_BLOCKING);
					return -1;
				}
				xc_set_errno(XN_ERR_RECV);
				return 0;
			}
			else if (!n)
			{
				x->mode = XN_DISCON;
				xc_set_errno(XN_ERR_CLOSED);
				return 0;
			}
			break;
		default:
			xc_set_errno(XN_ERR_UNSUPPORTED);
			return 0;
	}

	xc_set_errno(0);
	return n;
}

int xnet_recvs(xnet *x, char *buf, int len)
{
	int n;

	if ((n = xnet_recv(x, buf, len)) == -1)
		return 0;
	buf[n] = '\0';

	return 1;
}

const struct timeval null_timeval = {0, 0};
int xnet_newdata(xnet *x)
{
	switch (x->type)
	{
	case XN_TCP:
	case XN_TCP_BLOCKING:
	case XN_UDP:
		{
			fd_set fds;

			FD_ZERO(&fds);
			FD_SET(x->info.tcp.sock, &fds);

			return select(0, &fds, 0, 0, &null_timeval);
		}
	}
	return 0;
}

int xnet_cleanup()
{
	WSACleanup();

	return 1;
}

