#ifdef WIN32
#define _int16 short
#define _int32 int
#define _int64 __int64
#else
#define _int16 short
#define _int32 int
#define _int64 long long
#endif

#ifdef WIN32
	#include <windows.h>
	#include <winsock.h>
	#include <io.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <fcntl.h>

#define SIZE_8 0
#define SIZE_16 1
#define SIZE_32 2
#define SIZE_64 3

#define TYPE_CONST 0
#define TYPE_REG 4
#define TYPE_LPTR 8
#define TYPE_HPTR 12
#define TYPE_PTR 8

#define OP_NOP 0
#define OP_LD_RR 1
#define OP_LD_RC 5
#define OP_LD_RMC 9
#define OP_LD_RMR 13
#define OP_LD_RHC 17
#define OP_LD_RHR 21
#define OP_LD_MCR 25
#define OP_LD_MRR 29
#define OP_LD_HCR 33
#define OP_LD_HRR 37
#define OP_ADD 41
#define OP_FADD 45
#define OP_SUB 49
#define OP_FSUB 53
#define OP_MULT 57
#define OP_MULTS 61
#define OP_FMULT 65
#define OP_TEST 69
#define OP_DIV 73
#define OP_DIVS 77
#define OP_FDIV 81
#define OP_FCP 85
#define OP_MOD 89
#define OP_MODS 93
#define OP_SL 97
#define OP_SLN 101
#define OP_SR 105
#define OP_SRN 109
#define OP_NEG 113
#define OP_FNEG 117
#define OP_OR 121
#define OP_AND 125
#define OP_NOT 129
#define OP_XOR 133
#define OP_BOR 137
#define OP_BAND 141
#define OP_BNOT 145
#define OP_CP 149
#define OP_JP 153
#define OP_JP_Z 154
#define OP_JP_NZ 155
#define OP_JP_C 156
#define OP_JP_NC 157
#define OP_CALL_C 158
#define OP_CALL_R 159
#define OP_RET 160
#define OP_RETV 161
#define OP_HCALL_C 165
#define OP_HCALL_R 166
#define OP_INC 167
#define OP_DEC 171
#define OP_PUSH 175
#define OP_POP 179
#define OP_JP_R 183
#define OP_JP_R_Z 184
#define OP_JP_R_NZ 185
#define OP_JP_R_C 186
#define OP_JP_R_NC 187
#define OP_ARG_C 188
#define OP_ARG_R 192
#define OP_HCPY 196
#define OP_HSCPY 197
#define OP_MCPY 198
#define OP_CPY 199
#define OP_SCPY 200
#define OP_DCPY 201
#define OP_PMODE 202
#define OP_IMODE 203
#define OP_RMODE 204
#define OP_OMODE 205
#define OP_EMODE 206
#define OP_SYNC 207
#define OP_RSEND 208
#define OP_MSEND 209
#define OP_MBASE 210
#define OP_MALLOC 211
#define OP_FREE 212
#define OP_LALLOC 213
#define OP_LFREE 214
#define OP_LADDR 215
#define OP_SCP 216
#define OP_STEST 220
#define OP_SLEN 224
#define OP_CAST 228
#define OP_FINC 229
#define OP_FDEC 233
#define OP_GBF 237
#define OP_SBF 238
#define OP_SGBF 239
#define OP_SSBF 240
#define OP_ZERO 241
#define N_OPS 242

#define N_REG 16
#define S_REG 8
#define N_STACK 255
#define N_CSTACK 255
#define N_HOSTSYM 65535
#define N_ARG 255
#define MAX_DATA 255
#define MAX_DSZ 255
#define N_MODE 255
#define S_NBUF 255

#define XC_TYPE_VOID 0
#define XC_TYPE_CHAR 1
#define	XC_TYPE_INT 2
#define XC_TYPE_SHORT 4
#define XC_TYPE_LONG 5

#define ARGSZ 64
#define BS 1024
#define MAX_OUT 1048576
#define MAX_LNAME 255
#define MAX_CLINE 10
#define START_NARG 5

#define LOG_10_2 0.301029995664

typedef struct deft
{
	char name[MAX_LNAME], value[MAX_LNAME];
} deft;

typedef struct datat
{
	unsigned char *dat;
	int sz;
} datat;

typedef struct labelt
{
	char name[MAX_LNAME];
	unsigned _int32 loc;
	unsigned char filled;
} labelt;

typedef struct labelrt
{
	unsigned _int32 id;
	unsigned char *loc, sz;
} labelrt;

int mantissa[4] = {0, 0, 23, 52};
int exponent[4] = {0, 0, 8, 11};
int bias[4] = {0, 0, 127, 1023};
char tsz[4] = {1, 2, 4, 8};

